# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import functools
import os
from fnmatch import fnmatchcase
from typing import TYPE_CHECKING, Callable, Optional, TypedDict, Union, cast

import numpy as np
import pandas as pd

import mlrun
import mlrun.common.model_monitoring.helpers
import mlrun.common.schemas.model_monitoring.constants as mm_constants
import mlrun.data_types.infer
import mlrun.datastore.datastore_profile
import mlrun.platforms.iguazio
from mlrun.common.schemas import ModelEndpoint
from mlrun.common.schemas.model_monitoring.model_endpoints import (
    ModelEndpointMonitoringMetric,
    compose_full_name,
)
from mlrun.utils import logger

if TYPE_CHECKING:
    from mlrun.datastore import DataItem
    from mlrun.db.base import RunDBInterface
    from mlrun.projects import MlrunProject


class _BatchDict(TypedDict):
    minutes: int
    hours: int
    days: int


def _is_results_regex_match(
    existing_result_name: Optional[str],
    result_name_filters: Optional[list[str]],
) -> bool:
    if existing_result_name.count(".") != 3 or any(
        part == "" for part in existing_result_name.split(".")
    ):
        logger.warning(
            f"_is_results_regex_match: existing_result_name illegal, will be ignored."
            f" existing_result_name: {existing_result_name}"
        )
        return False
    existing_result_name = ".".join(existing_result_name.split(".")[i] for i in [1, 3])
    for result_name_filter in result_name_filters:
        if fnmatchcase(existing_result_name, result_name_filter):
            return True
    return False


def filter_results_by_regex(
    existing_result_names: Optional[list[str]] = None,
    result_name_filters: Optional[list[str]] = None,
) -> list[str]:
    """
    Filter a list of existing result names by a list of filters.

    This function returns only the results that match the filters provided. If no filters are given,
    it returns all results. Invalid inputs are ignored.

    :param existing_result_names: List of existing results' fully qualified names (FQNs)
                                  in the format: endpoint_id.app_name.type.name.
                                  Example: mep1.app1.result.metric1
    :param result_name_filters:   List of filters in the format: app.result_name.
                                  Wildcards can be used, such as app.result* or *.result

    :return: List of FQNs of the matching results
    """

    if not result_name_filters:
        return existing_result_names

    if not existing_result_names:
        return []

    #  filters validations
    validated_filters = []
    for result_name_filter in result_name_filters:
        if result_name_filter.count(".") != 1:
            logger.warning(
                f"filter_results_by_regex: result_name_filter illegal, will be ignored."
                f"Filter: {result_name_filter}"
            )
        else:
            validated_filters.append(result_name_filter)
    filtered_metrics_names = []
    for existing_result_name in existing_result_names:
        if _is_results_regex_match(
            existing_result_name=existing_result_name,
            result_name_filters=validated_filters,
        ):
            filtered_metrics_names.append(existing_result_name)
    return list(set(filtered_metrics_names))


def get_stream_path(
    project: str,
    function_name: str = mm_constants.MonitoringFunctionNames.STREAM,
    stream_uri: Optional[str] = None,
    secret_provider: Optional[Callable[[str], str]] = None,
    profile: Optional[mlrun.datastore.datastore_profile.DatastoreProfile] = None,
) -> str:
    """
    Get stream path from the project secret. If wasn't set, take it from the system configurations

    :param project:             Project name.
    :param function_name:       Application name. Default is model_monitoring_stream.
    :param stream_uri:          Stream URI. If provided, it will be used instead of the one from the project's secret.
    :param secret_provider:     Optional secret provider to get the connection string secret.
                                If not set, the env vars are used.
    :param profile:             Optional datastore profile of the stream (V3IO/KafkaSource profile).
    :return:                    Monitoring stream path to the relevant application.
    """

    profile = profile or _get_stream_profile(
        project=project, secret_provider=secret_provider
    )

    if isinstance(profile, mlrun.datastore.datastore_profile.DatastoreProfileV3io):
        stream_uri = mlrun.mlconf.get_model_monitoring_file_target_path(
            project=project,
            kind=mm_constants.FileTargetKind.STREAM,
            target="online",
            function_name=function_name,
        )
        return stream_uri.replace("v3io://", f"ds://{profile.name}")

    elif isinstance(
        profile, mlrun.datastore.datastore_profile.DatastoreProfileKafkaStream
    ):
        topic = mlrun.common.model_monitoring.helpers.get_kafka_topic(
            project=project, function_name=function_name
        )
        return f"ds://{profile.name}/{topic}"
    else:
        raise mlrun.errors.MLRunValueError(
            f"Received an unexpected stream profile type: {type(profile)}\n"
            "Expects `DatastoreProfileV3io` or `DatastoreProfileKafkaStream`."
        )


def get_monitoring_parquet_path(
    project: "MlrunProject",
    kind: str = mm_constants.FileTargetKind.PARQUET,
) -> str:
    """Get model monitoring parquet target for the current project and kind. The parquet target path is based on the
    project artifact path. If project artifact path is not defined, the parquet target path will be based on MLRun
    artifact path.

    :param project:     Project object.
    :param kind:        indicate the kind of the parquet path, can be either stream_parquet or stream_controller_parquet

    :return:           Monitoring parquet target path.
    """
    artifact_path = project.spec.artifact_path
    # Generate monitoring parquet path value
    parquet_path = mlrun.mlconf.get_model_monitoring_file_target_path(
        project=project.name,
        kind=kind,
        target="offline",
        artifact_path=artifact_path,
    )
    return parquet_path


def get_monitoring_stats_directory_path(
    project: str,
    kind: str = mm_constants.FileTargetKind.STATS,
) -> str:
    """
    Get model monitoring stats target for the current project and kind. The stats target path is based on the
    project artifact path. If project artifact path is not defined, the stats target path will be based on MLRun
    artifact path.
    :param project:     Project object.
    :param kind:        indicate the kind of the stats path
    :return:            Monitoring stats target path.
    """
    stats_path = mlrun.mlconf.get_model_monitoring_file_target_path(
        project=project,
        kind=kind,
    )
    return stats_path


def _get_monitoring_current_stats_file_path(project: str, endpoint_id: str) -> str:
    return os.path.join(
        get_monitoring_stats_directory_path(project),
        f"{endpoint_id}_current_stats.json",
    )


def _get_monitoring_drift_measures_file_path(project: str, endpoint_id: str) -> str:
    return os.path.join(
        get_monitoring_stats_directory_path(project),
        f"{endpoint_id}_drift_measures.json",
    )


def get_monitoring_current_stats_data(project: str, endpoint_id: str) -> "DataItem":
    """
    getter for data item of current stats for project and endpoint
    :param project: project name str
    :param endpoint_id: endpoint id str
    :return: DataItem
    """
    return mlrun.datastore.store_manager.object(
        _get_monitoring_current_stats_file_path(
            project=project, endpoint_id=endpoint_id
        )
    )


def get_monitoring_drift_measures_data(project: str, endpoint_id: str) -> "DataItem":
    """
    getter for data item of drift measures for project and endpoint
    :param project: project name str
    :param endpoint_id: endpoint id str
    :return: DataItem
    """
    return mlrun.datastore.store_manager.object(
        _get_monitoring_drift_measures_file_path(
            project=project, endpoint_id=endpoint_id
        )
    )


def _get_profile(
    project: str,
    secret_provider: Optional[Callable[[str], str]],
    profile_name_key: str,
) -> mlrun.datastore.datastore_profile.DatastoreProfile:
    """
    Get the datastore profile from the project name and secret provider, where the profile's name
    is saved as a secret named `profile_name_key`.

    :param project:          The project name.
    :param secret_provider:  Secret provider to get the secrets from, or `None` for env vars.
    :param profile_name_key: The profile name key in the secret store.
    :return:                 Datastore profile.
    """
    profile_name = mlrun.get_secret_or_env(
        key=profile_name_key, secret_provider=secret_provider
    )
    if not profile_name:
        raise mlrun.errors.MLRunNotFoundError(
            f"Not found `{profile_name_key}` profile name for project '{project}'"
        )
    return mlrun.datastore.datastore_profile.datastore_profile_read(
        url=f"ds://{profile_name}", project_name=project, secrets=secret_provider
    )


_get_tsdb_profile = functools.partial(
    _get_profile, profile_name_key=mm_constants.ProjectSecretKeys.TSDB_PROFILE_NAME
)
_get_stream_profile = functools.partial(
    _get_profile, profile_name_key=mm_constants.ProjectSecretKeys.STREAM_PROFILE_NAME
)


def _get_v3io_output_stream(
    *,
    v3io_profile: mlrun.datastore.datastore_profile.DatastoreProfileV3io,
    project: str,
    function_name: str,
    v3io_access_key: Optional[str],
    mock: bool = False,
) -> mlrun.platforms.iguazio.OutputStream:
    stream_uri = mlrun.mlconf.get_model_monitoring_file_target_path(
        project=project,
        kind=mm_constants.FileTargetKind.STREAM,
        target="online",
        function_name=function_name,
    )
    endpoint, stream_path = mlrun.platforms.iguazio.parse_path(stream_uri)
    return mlrun.platforms.iguazio.OutputStream(
        stream_path,
        endpoint=endpoint,
        access_key=v3io_access_key or v3io_profile.v3io_access_key,
        mock=mock,
    )


def _get_kafka_output_stream(
    *,
    kafka_profile: mlrun.datastore.datastore_profile.DatastoreProfileKafkaStream,
    project: str,
    function_name: str,
    mock: bool = False,
) -> mlrun.platforms.iguazio.KafkaOutputStream:
    topic = mlrun.common.model_monitoring.helpers.get_kafka_topic(
        project=project, function_name=function_name
    )
    attributes = kafka_profile.attributes()
    producer_options = mlrun.datastore.utils.KafkaParameters(attributes).producer()

    return mlrun.platforms.iguazio.KafkaOutputStream(
        brokers=kafka_profile.brokers,
        topic=topic,
        producer_options=producer_options,
        mock=mock,
    )


def get_output_stream(
    project: str,
    function_name: str = mm_constants.MonitoringFunctionNames.STREAM,
    secret_provider: Optional[Callable[[str], str]] = None,
    profile: Optional[mlrun.datastore.datastore_profile.DatastoreProfile] = None,
    v3io_access_key: Optional[str] = None,
    mock: bool = False,
) -> Union[
    mlrun.platforms.iguazio.OutputStream, mlrun.platforms.iguazio.KafkaOutputStream
]:
    """
    Get stream path from the project secret. If wasn't set, take it from the system configurations

    :param project:             Project name.
    :param function_name:       Application name. Default is model_monitoring_stream.
    :param secret_provider:     Optional secret provider to get the connection string secret.
                                If not set, the env vars are used.
    :param profile:             Optional datastore profile of the stream (V3IO/KafkaSource profile).
    :param v3io_access_key:     Optional V3IO access key.
    :param mock:                Should the output stream be mocked or not.
    :return:                    Monitoring stream path to the relevant application.
    """

    profile = profile or _get_stream_profile(
        project=project, secret_provider=secret_provider
    )

    if isinstance(profile, mlrun.datastore.datastore_profile.DatastoreProfileV3io):
        return _get_v3io_output_stream(
            v3io_profile=profile,
            project=project,
            function_name=function_name,
            v3io_access_key=v3io_access_key,
            mock=mock,
        )

    elif isinstance(
        profile, mlrun.datastore.datastore_profile.DatastoreProfileKafkaStream
    ):
        return _get_kafka_output_stream(
            kafka_profile=profile,
            project=project,
            function_name=function_name,
            mock=mock,
        )

    else:
        raise mlrun.errors.MLRunValueError(
            f"Received an unexpected stream profile type: {type(profile)}\n"
            "Expects `DatastoreProfileV3io` or `DatastoreProfileKafkaStream`."
        )


def batch_dict2timedelta(batch_dict: _BatchDict) -> datetime.timedelta:
    """
    Convert a batch dictionary to timedelta.

    :param batch_dict:  Batch dict.

    :return:            Timedelta.
    """
    return datetime.timedelta(**batch_dict)


def _get_monitoring_time_window_from_controller_run(
    project: str, db: "RunDBInterface"
) -> datetime.timedelta:
    """
    Get the base period form the controller.

    :param project: Project name.
    :param db:      DB interface.

    :return:    Timedelta for the controller to run.
    :raise:     MLRunNotFoundError if the controller isn't deployed yet
    """

    controller = db.get_function(
        name=mm_constants.MonitoringFunctionNames.APPLICATION_CONTROLLER,
        project=project,
    )
    if isinstance(controller, dict):
        controller = mlrun.runtimes.RemoteRuntime.from_dict(controller)
    elif not hasattr(controller, "to_dict"):
        raise mlrun.errors.MLRunNotFoundError()
    base_period = controller.spec.config["spec.triggers.cron_interval"]["attributes"][
        "interval"
    ]
    batch_dict = {
        mm_constants.EventFieldType.MINUTES: int(base_period[:-1]),
        mm_constants.EventFieldType.HOURS: 0,
        mm_constants.EventFieldType.DAYS: 0,
    }
    return batch_dict2timedelta(batch_dict)


def update_model_endpoint_last_request(
    project: str,
    model_endpoint: ModelEndpoint,
    current_request: datetime.datetime,
    db: "RunDBInterface",
) -> None:
    """
    Update the last request field of the model endpoint to be after the current request time.

    :param project:         Project name.
    :param model_endpoint:  Model endpoint object.
    :param current_request: current request time
    :param db:              DB interface.
    """

    logger.info(
        "Update model endpoint last request time (EP with serving)",
        project=project,
        endpoint_id=model_endpoint.metadata.uid,
        name=model_endpoint.metadata.name,
        function_name=model_endpoint.spec.function_name,
        last_request=model_endpoint.status.last_request,
        current_request=current_request,
    )
    db.patch_model_endpoint(
        project=project,
        endpoint_id=model_endpoint.metadata.uid,
        name=model_endpoint.metadata.name,
        function_name=model_endpoint.spec.function_name,
        attributes={mm_constants.EventFieldType.LAST_REQUEST: current_request},
    )


def calculate_inputs_statistics(
    sample_set_statistics: dict, inputs: pd.DataFrame
) -> mlrun.common.model_monitoring.helpers.FeatureStats:
    """
    Calculate the inputs data statistics for drift monitoring purpose.

    :param sample_set_statistics: The sample set (stored end point's dataset to reference) statistics. The bins of the
                                  histograms of each feature will be used to recalculate the histograms of the inputs.
    :param inputs:                The inputs to calculate their statistics and later on - the drift with respect to the
                                  sample set.

    :returns: The calculated statistics of the inputs data.
    """

    # Use `DFDataInfer` to calculate the statistics over the inputs:
    inputs_statistics = mlrun.data_types.infer.DFDataInfer.get_stats(
        df=inputs, options=mlrun.data_types.infer.InferOptions.Histogram
    )

    # Recalculate the histograms over the bins that are set in the sample-set of the end point:
    for feature in list(inputs_statistics):
        if feature in sample_set_statistics:
            counts, bins = np.histogram(
                inputs[feature].to_numpy(),
                bins=sample_set_statistics[feature]["hist"][1],
            )
            inputs_statistics[feature]["hist"] = [
                counts.tolist(),
                bins.tolist(),
            ]
        else:
            # If the feature is not in the sample set and doesn't have a histogram, remove it from the statistics:
            inputs_statistics.pop(feature)

    return inputs_statistics


def get_result_instance_fqn(
    model_endpoint_id: str, app_name: str, result_name: str
) -> str:
    return f"{model_endpoint_id}.{app_name}.result.{result_name}"


def get_alert_name_from_result_fqn(result_fqn: str):
    """
    :param   result_fqn: current get_result_instance_fqn format: `{model_endpoint_id}.{app_name}.result.{result_name}`

    :return: shorter fqn without forbidden alert characters.
    """
    if result_fqn.count(".") != 3 or result_fqn.split(".")[2] != "result":
        raise mlrun.errors.MLRunValueError(
            f"result_fqn: {result_fqn} is not in the correct format: {{model_endpoint_id}}.{{app_name}}."
            f"result.{{result_name}}"
        )
    # Name format cannot contain "."
    # The third component is always `result`, so it is not necessary for checking uniqueness.
    return "_".join(result_fqn.split(".")[i] for i in [0, 1, 3])


def get_default_result_instance_fqn(model_endpoint_id: str) -> str:
    return get_result_instance_fqn(
        model_endpoint_id,
        mm_constants.HistogramDataDriftApplicationConstants.NAME,
        mm_constants.HistogramDataDriftApplicationConstants.GENERAL_RESULT_NAME,
    )


def get_invocations_fqn(project: str) -> str:
    return compose_full_name(
        project=project,
        app=mm_constants.SpecialApps.MLRUN_INFRA,
        name=mm_constants.PredictionsQueryConstants.INVOCATIONS,
        type=mm_constants.ModelEndpointMonitoringMetricType.METRIC,
    )


def get_invocations_metric(project: str) -> ModelEndpointMonitoringMetric:
    """
    Return the invocations metric of any model endpoint in the given project.

    :param project: The project name.
    :returns:       The model monitoring metric object.
    """
    return ModelEndpointMonitoringMetric(
        project=project,
        app=mm_constants.SpecialApps.MLRUN_INFRA,
        type=mm_constants.ModelEndpointMonitoringMetricType.METRIC,
        name=mm_constants.PredictionsQueryConstants.INVOCATIONS,
        full_name=get_invocations_fqn(project),
    )


def _get_monitoring_schedules_folder_path(project: str) -> str:
    return cast(
        str,
        mlrun.mlconf.get_model_monitoring_file_target_path(
            project=project, kind=mm_constants.FileTargetKind.MONITORING_SCHEDULES
        ),
    )


def _get_monitoring_schedules_user_folder_path(out_path: str) -> str:
    return os.path.join(out_path, mm_constants.FileTargetKind.MONITORING_SCHEDULES)


def _get_monitoring_schedules_file_endpoint_path(
    *, project: str, endpoint_id: str
) -> str:
    return os.path.join(
        _get_monitoring_schedules_folder_path(project), f"{endpoint_id}.json"
    )


def get_monitoring_schedules_endpoint_data(
    *, project: str, endpoint_id: str
) -> "DataItem":
    """
    Get the model monitoring schedules' data item of the project's model endpoint.
    """
    return mlrun.datastore.store_manager.object(
        _get_monitoring_schedules_file_endpoint_path(
            project=project, endpoint_id=endpoint_id
        )
    )


def get_monitoring_schedules_chief_data(*, project: str) -> "DataItem":
    """
    Get the model monitoring schedules' data item of the project's model endpoint.
    """
    return mlrun.datastore.store_manager.object(
        _get_monitoring_schedules_file_chief_path(project=project)
    )


def get_monitoring_schedules_user_application_data(
    *, out_path: str, application: str
) -> "DataItem":
    """
    Get the model monitoring schedules' data item of user application runs.
    """
    return mlrun.datastore.store_manager.object(
        _get_monitoring_schedules_file_user_application_path(
            out_path=out_path, application=application
        )
    )


def _get_monitoring_schedules_file_chief_path(
    *,
    project: str,
) -> str:
    return os.path.join(
        _get_monitoring_schedules_folder_path(project), f"{project}.json"
    )


def _get_monitoring_schedules_file_user_application_path(
    *, out_path: str, application: str
) -> str:
    return os.path.join(
        _get_monitoring_schedules_user_folder_path(out_path), f"{application}.json"
    )


def get_start_end(
    start: Union[datetime.datetime, None],
    end: Union[datetime.datetime, None],
    delta: Optional[datetime.timedelta] = None,
) -> tuple[datetime.datetime, datetime.datetime]:
    """
    static utils function for tsdb start end format
    :param start:       Either None or datetime, None is handled as datetime.min(tz=timezone.utc) unless `delta`
                        is provided.
    :param end:         Either None or datetime, None is handled as datetime.now(tz=timezone.utc)
    :param delta:       Optional timedelta to define a time span.
                        - If both `start` and `end` are provided, `delta` is ignored.
                        - If only one of `start` or `end` is provided, the other will be
                          calculated using `delta`.
                        - If neither `start` nor `end` is provided, `end` defaults to now,
                          and `start` is calculated as `end - delta`.
    :return:            start datetime, end datetime
    """

    if delta and start and end:
        # If both start and end are provided, delta is ignored
        pass
    elif delta:
        if start and not end:
            end = start + delta
        else:
            end = end or mlrun.utils.datetime_now()
            start = end - delta
    else:
        start = start or mlrun.utils.datetime_min()
        end = end or mlrun.utils.datetime_now()

    if not (
        isinstance(start, datetime.datetime) and isinstance(end, datetime.datetime)
    ):
        raise mlrun.errors.MLRunInvalidArgumentError(
            "Both start and end must be datetime objects"
        )

    if start > end:
        raise mlrun.errors.MLRunInvalidArgumentError(
            "The start time must be before the end time. Note that if end time is not provided, "
            "the current time is used by default"
        )

    return start, end


def validate_time_range(
    start: Optional[datetime.datetime] = None, end: Optional[datetime.datetime] = None
) -> tuple[datetime.datetime, datetime.datetime]:
    """
    validate start and end parameters and set default values if needed.
    :param start:       Either None or datetime, None is handled as datetime.now(tz=timezone.utc) - timedelta(days=1)
    :param end:         Either None or datetime, None is handled as datetime.now(tz=timezone.utc)
    :return:            start datetime, end datetime
    """
    end = end or mlrun.utils.helpers.datetime_now()
    start = start or (end - datetime.timedelta(days=1))
    if start.tzinfo is None or end.tzinfo is None:
        raise mlrun.errors.MLRunInvalidArgumentTypeError(
            "Custom start and end times must contain the timezone."
        )
    if start > end:
        raise mlrun.errors.MLRunInvalidArgumentError(
            "The start time must be before the end time. Note that if end time is not provided, "
            "the current time is used by default."
        )
    return start, end



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import enum
import typing
from typing import Optional

import pydantic.v1

import mlrun.common.types


class NotificationKind(mlrun.common.types.StrEnum):
    """Currently, the supported notification kinds and their params are as follows:"""

    console: str = "console"
    """no params, local only"""

    git: str = "git"
    """
    **token** - The git token to use for the git notification.\n
    **repo** - The git repo to which to send the notification.\n
    **issue** - The git issue to which to send the notification.\n
    **merge_request** -
     In GitLab (as opposed to GitHub), merge requests and issues are separate entities.
     If using merge request, the issue will be ignored, and vice versa.\n
    **server** - The git server to which to send the notification.\n
    **gitlab** - (bool) Whether the git server is GitLab or not.\n
    """

    ipython: str = "ipython"
    """no params, local only"""

    slack: str = "slack"
    """**webhook** - The slack webhook to which to send the notification."""

    mail: str = "mail"
    """
    **email_addresses** - The target mails\n
    **subject** - The subject of the mail\n
    **body** - The body of the mail\n
    """

    webhook: str = "webhook"
    """
    **url** - The webhook url to which to send the notification.\n
    **method** - The http method to use when sending the notification (GET, POST, PUT, etc…).\n
    **headers** - (dict) The http headers to send with the notification.\n
    **override_body** -
     (dict) The body to send with the notification. If not specified, the
     default body will be a dictionary containing `name`, `message`, `severity`, and a `runs` list of the
     completed runs. You can also add the run's details.\n
     Example::

                "override_body": {"message":"Run Completed {{ runs }}"
                # Results would look like:
                "message": "Run Completed [{'project': 'my-project', 'name': 'my-function', 'host': <run-host>,
                         'status': {'state': 'completed', 'results': <run-results>}}]"
    **verify_ssl** -
     (bool) Whether SSL certificates are validated during HTTP requests or not.
     The default is set to True.\n
    """

    @classmethod
    def alert_notification_kinds(cls) -> list[str]:
        return [cls.git, cls.slack, cls.webhook]


class NotificationSeverity(mlrun.common.types.StrEnum):
    INFO = "info"
    DEBUG = "debug"
    VERBOSE = "verbose"
    WARNING = "warning"
    ERROR = "error"


class NotificationStatus(mlrun.common.types.StrEnum):
    PENDING = "pending"
    SENT = "sent"
    ERROR = "error"


class NotificationLimits(enum.Enum):
    max_params_size = (
        900 * 1024
    )  # 900KB (k8s secret size limit is 1MB minus buffer for metadata)


class Notification(pydantic.v1.BaseModel):
    """
    Notification object schema

    :param kind: notification implementation kind - slack, webhook, etc.
    :param name: for logging and identification
    :param message: message content in the notification
    :param severity: severity to display in the notification
    :param when: list of statuses to trigger the notification: 'running', 'completed', 'error'
    :param condition: optional condition to trigger the notification, a jinja2 expression that can use run data
                      to evaluate if the notification should be sent in addition to the 'when' statuses.
                      e.g.: '{{ run["status"]["results"]["accuracy"] < 0.9}}'
    :param params: Implementation specific parameters for the notification implementation (e.g. slack webhook url,
                   git repository details, etc.)
    :param secret_params: secret parameters for the notification implementation, same as params but will be stored
                          in a k8s secret and passed as a secret reference to the implementation.
    :param status: notification status - pending, sent, error
    :param sent_time: time the notification was sent
    :param reason: failure reason if the notification failed to send
    """

    kind: NotificationKind
    name: str
    message: typing.Optional[str] = None
    severity: typing.Optional[NotificationSeverity] = None
    when: typing.Optional[list[str]] = None
    condition: typing.Optional[str] = None
    params: typing.Optional[dict[str, typing.Any]] = None
    status: typing.Optional[NotificationStatus] = None
    sent_time: typing.Optional[typing.Union[str, datetime.datetime]] = None
    secret_params: typing.Optional[dict[str, typing.Any]] = None
    reason: typing.Optional[str] = None


class SetNotificationRequest(pydantic.v1.BaseModel):
    notifications: list[Notification] = None


class NotificationSummary(pydantic.v1.BaseModel):
    failed: int = 0
    succeeded: int = 0


class NotificationState(pydantic.v1.BaseModel):
    kind: str
    err: Optional[
        str
    ]  # empty error means that the notifications were sent successfully
    summary: NotificationSummary



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
import tempfile
from typing import Optional, Union

from mlrun.artifacts import Artifact
from mlrun.datastore import DataItem
from mlrun.errors import MLRunInvalidArgumentError

from ..utils import (
    DEFAULT_ARCHIVE_FORMAT,
    DEFAULT_STRUCT_FILE_FORMAT,
    ArchiveSupportedFormat,
    ArtifactType,
    StructFileSupportedFormat,
)
from .default_packager import DefaultPackager

# ----------------------------------------------------------------------------------------------------------------------
# builtins packagers:
# ----------------------------------------------------------------------------------------------------------------------


class NonePackager(DefaultPackager):
    """
    ``None`` packager.
    """

    # TODO: From python 3.10 the `PACKABLE_OBJECT_TYPE` should be changed to `types.NoneType`
    PACKABLE_OBJECT_TYPE = type(None)
    DEFAULT_PACKING_ARTIFACT_TYPE = ArtifactType.RESULT

    # TODO: `None` as pickle will be available from Python 3.10, so this method can be removed once we move to 3.10.
    def get_supported_artifact_types(self) -> list[str]:
        """
        Get all the supported artifact types on this packager. It will be the same as `DefaultPackager` but without the
        'object' artifact type support (None cannot be pickled, only from Python 3.10, and it should not be pickled
        anyway as it is simply None - a result will do).

        :return: A list of all the supported artifact types.
        """
        supported_artifacts = super().get_supported_artifact_types()
        supported_artifacts.remove("object")
        return supported_artifacts


class IntPackager(DefaultPackager):
    """
    ``builtins.int`` packager.
    """

    PACKABLE_OBJECT_TYPE = int
    DEFAULT_PACKING_ARTIFACT_TYPE = ArtifactType.RESULT


class FloatPackager(DefaultPackager):
    """
    ``builtins.float`` packager.
    """

    PACKABLE_OBJECT_TYPE = float
    DEFAULT_PACKING_ARTIFACT_TYPE = ArtifactType.RESULT


class BoolPackager(DefaultPackager):
    """
    ``builtins.bool`` packager.
    """

    PACKABLE_OBJECT_TYPE = bool
    DEFAULT_PACKING_ARTIFACT_TYPE = ArtifactType.RESULT


class StrPackager(DefaultPackager):
    """
    ``builtins.str`` packager.
    """

    PACKABLE_OBJECT_TYPE = str
    DEFAULT_PACKING_ARTIFACT_TYPE = ArtifactType.RESULT
    DEFAULT_UNPACKING_ARTIFACT_TYPE = ArtifactType.PATH

    def pack_path(
        self, obj: str, key: str, archive_format: str = DEFAULT_ARCHIVE_FORMAT
    ) -> tuple[Artifact, dict]:
        """
        Pack a path string value content (pack the file or directory in that path).

        :param obj:            The string path value to pack.
        :param key:            The key to use for the artifact.
        :param archive_format: The archive format to use in case the path is of a directory. Default is zip.

        :return: The packed artifact and instructions.
        """
        # TODO: Add a configuration like `archive_file: bool = False` to enable archiving a single file to shrink it in
        #       size. In that case the `is_directory` instruction will make it so when an archive is received, if its
        #       a directory, when exporting it a directory path should be returned. And, if its a file, a path to the
        #       single file exported should be returned.
        # Verify the path is of an existing file:
        if not os.path.exists(obj):
            raise MLRunInvalidArgumentError(f"The given path do not exist: '{obj}'")

        # Proceed by path type (file or directory):
        if os.path.isfile(obj):
            # Create the artifact:
            artifact = Artifact(key=key, src_path=os.path.abspath(obj))
            instructions = {"is_directory": False}
        elif os.path.isdir(obj):
            # Archive the directory:
            output_path = tempfile.mkdtemp()
            archiver = ArchiveSupportedFormat.get_format_handler(fmt=archive_format)
            archive_path = archiver.create_archive(
                directory_path=obj, output_path=output_path
            )
            # Create the artifact:
            artifact = Artifact(key=key, src_path=archive_path)
            instructions = {"archive_format": archive_format, "is_directory": True}
        else:
            raise MLRunInvalidArgumentError(
                f"The given path is not a file nor a directory: '{obj}'"
            )

        return artifact, instructions

    def unpack_path(
        self,
        data_item: DataItem,
        is_directory: bool = False,
        archive_format: Optional[str] = None,
    ) -> str:
        """
        Unpack a data item representing a path string. If the path is of a file, the file is downloaded to a local
        temporary directory and its path is returned. If the path is of a directory, the archive is extracted and the
        directory path extracted is returned.

        :param data_item:      The data item to unpack.
        :param is_directory:   Whether the path should be treated as a file or a directory. Files (even archives like
                               zip) won't be extracted.
        :param archive_format: The archive format to use in case the path is of a directory. Default is None - will be
                               read by the archive file extension.

        :return: The unpacked string.
        """
        # Get the file:
        path = self.get_data_item_local_path(data_item=data_item)

        # If it's not a directory, return the file path. Otherwise, it should be extracted according to the archive
        # format:
        if not is_directory:
            return path

        # Get the archive format by the file extension:
        if archive_format is None:
            archive_format = ArchiveSupportedFormat.match_format(path=path)
        if archive_format is None:
            raise MLRunInvalidArgumentError(
                f"Archive format of {data_item.key} ('{''.join(pathlib.Path(path).suffixes)}') is not supported. "
                f"Supported formats are: {' '.join(ArchiveSupportedFormat.get_all_formats())}"
            )

        # Extract the archive:
        archiver = ArchiveSupportedFormat.get_format_handler(fmt=archive_format)
        directory_path = archiver.extract_archive(
            archive_path=path, output_path=os.path.dirname(path)
        )

        # Mark the extracted content for future clear:
        self.add_future_clearing_path(path=directory_path)

        # Return the extracted directory path:
        return directory_path


class _BuiltinCollectionPackager(DefaultPackager):
    """
    A base packager for builtin python dictionaries and lists as they share common artifact and file types.
    """

    DEFAULT_PACKING_ARTIFACT_TYPE = ArtifactType.RESULT
    DEFAULT_UNPACKING_ARTIFACT_TYPE = ArtifactType.FILE

    def pack_file(
        self,
        obj: Union[dict, list],
        key: str,
        file_format: str = DEFAULT_STRUCT_FILE_FORMAT,
    ) -> tuple[Artifact, dict]:
        """
        Pack a builtin collection as a file by the given format.

        :param obj:         The builtin collection to pack.
        :param key:         The key to use for the artifact.
        :param file_format: The file format to save as. Default is json.

        :return: The packed artifact and instructions.
        """
        # Write to file:
        formatter = StructFileSupportedFormat.get_format_handler(fmt=file_format)
        temp_directory = pathlib.Path(tempfile.mkdtemp())
        self.add_future_clearing_path(path=temp_directory)
        file_path = temp_directory / f"{key}.{file_format}"
        formatter.write(obj=obj, file_path=str(file_path))

        # Create the artifact and instructions:
        artifact = Artifact(key=key, src_path=os.path.abspath(file_path))
        instructions = {"file_format": file_format}

        return artifact, instructions

    def unpack_file(
        self, data_item: DataItem, file_format: Optional[str] = None
    ) -> Union[dict, list]:
        """
        Unpack a builtin collection from file.

        :param data_item:   The data item to unpack.
        :param file_format: The file format to use for reading the builtin collection. Default is None - will be read by
                            the file extension.

        :return: The unpacked builtin collection.
        """
        # Get the file:
        file_path = self.get_data_item_local_path(data_item=data_item)

        # Get the archive format by the file extension if needed:
        if file_format is None:
            file_format = StructFileSupportedFormat.match_format(path=file_path)
        if file_format is None:
            raise MLRunInvalidArgumentError(
                f"File format of {data_item.key} ('{''.join(pathlib.Path(file_path).suffixes)}') is not supported. "
                f"Supported formats are: {' '.join(StructFileSupportedFormat.get_all_formats())}"
            )

        # Read the object:
        formatter = StructFileSupportedFormat.get_format_handler(fmt=file_format)
        obj = formatter.read(file_path=file_path)

        return obj


class DictPackager(_BuiltinCollectionPackager):
    """
    ``builtins.dict`` packager.
    """

    PACKABLE_OBJECT_TYPE = dict

    def unpack_file(
        self, data_item: DataItem, file_format: Optional[str] = None
    ) -> dict:
        """
        Unpack a dictionary from file.

        :param data_item:   The data item to unpack.
        :param file_format: The file format to use for reading the dictionary. Default is None - will be read by the
                            file extension.

        :return: The unpacked dictionary.
        """
        # Unpack the object:
        obj = super().unpack_file(data_item=data_item, file_format=file_format)

        # Check if needed to cast from list:
        if isinstance(obj, list):
            return {index: element for index, element in enumerate(obj)}
        return obj


class ListPackager(_BuiltinCollectionPackager):
    """
    ``builtins.list`` packager.
    """

    PACKABLE_OBJECT_TYPE = list

    def unpack_file(
        self, data_item: DataItem, file_format: Optional[str] = None
    ) -> list:
        """
        Unpack a list from file.

        :param data_item:   The data item to unpack.
        :param file_format: The file format to use for reading the list. Default is None - will be read by the file
                            extension.

        :return: The unpacked list.
        """
        # Unpack the object:
        obj = super().unpack_file(data_item=data_item, file_format=file_format)

        # Check if needed to cast from dict:
        if isinstance(obj, dict):
            return list(obj.values())
        return obj


class TuplePackager(ListPackager):
    """
    ``builtins.tuple`` packager.

    Notice: a ``tuple`` returned from a function is usually treated as multiple returned objects, and so MLRun will try
    to pack each of them separately and not as a single tuple. For example::

        def example_func_1():
            return 10, [1, 2, 3], "Hello MLRun"

    Will be returned as a ``tuple`` of 3 items: `(10, [1, 2, 3], "Hello MLRun")` but the items will be packaged
    separately one by one and not as a single ``tuple``.

    In order to pack tuples (not recommended), use the configuration::

        mlrun.mlconf.packagers.pack_tuple = True

    Or more correctly, cast your returned tuple to a ``list`` like so::

        def example_func_2():
            my_tuple = (2, 4)
            return list(my_tuple)
    """

    PACKABLE_OBJECT_TYPE = tuple

    def pack_result(self, obj: tuple, key: str) -> dict:
        """
        Pack a tuple as a result.

        :param obj: The tuple to pack and log.
        :param key: The result's key.

        :return: The result dictionary.
        """
        return super().pack_result(obj=list(obj), key=key)

    def pack_file(
        self, obj: tuple, key: str, file_format: str = DEFAULT_STRUCT_FILE_FORMAT
    ) -> tuple[Artifact, dict]:
        """
        Pack a tuple as a file by the given format.

        :param obj:         The tuple to pack.
        :param key:         The key to use for the artifact.
        :param file_format: The file format to save as. Default is json.

        :return: The packed artifact and instructions.
        """
        return super().pack_file(obj=list(obj), key=key, file_format=file_format)

    def unpack_file(
        self, data_item: DataItem, file_format: Optional[str] = None
    ) -> tuple:
        """
        Unpack a tuple from file.

        :param data_item:   The data item to unpack.
        :param file_format: The file format to use for reading the tuple. Default is None - will be read by the file
                            extension.

        :return: The unpacked tuple.
        """
        return tuple(super().unpack_file(data_item=data_item, file_format=file_format))


class SetPackager(ListPackager):
    """
    ``builtins.set`` packager.
    """

    PACKABLE_OBJECT_TYPE = set

    def pack_result(self, obj: set, key: str) -> dict:
        """
        Pack a set as a result.

        :param obj: The set to pack and log.
        :param key: The result's key.

        :return: The result dictionary.
        """
        return super().pack_result(obj=list(obj), key=key)

    def pack_file(
        self, obj: set, key: str, file_format: str = DEFAULT_STRUCT_FILE_FORMAT
    ) -> tuple[Artifact, dict]:
        """
        Pack a set as a file by the given format.

        :param obj:         The set to pack.
        :param key:         The key to use for the artifact.
        :param file_format: The file format to save as. Default is json.

        :return: The packed artifact and instructions.
        """
        return super().pack_file(obj=list(obj), key=key, file_format=file_format)

    def unpack_file(
        self, data_item: DataItem, file_format: Optional[str] = None
    ) -> set:
        """
        Unpack a set from file.

        :param data_item:   The data item to unpack.
        :param file_format: The file format to use for reading the set. Default is None - will be read by the file
                            extension.

        :return: The unpacked set.
        """
        return set(super().unpack_file(data_item=data_item, file_format=file_format))


class FrozensetPackager(SetPackager):
    """
    ``builtins.frozenset`` packager.
    """

    PACKABLE_OBJECT_TYPE = frozenset

    def pack_file(
        self, obj: frozenset, key: str, file_format: str = DEFAULT_STRUCT_FILE_FORMAT
    ) -> tuple[Artifact, dict]:
        """
        Pack a frozenset as a file by the given format.

        :param obj:         The frozenset to pack.
        :param key:         The key to use for the artifact.
        :param file_format: The file format to save as. Default is json.

        :return: The packed artifact and instructions.
        """
        return super().pack_file(obj=set(obj), key=key, file_format=file_format)

    def unpack_file(
        self, data_item: DataItem, file_format: Optional[str] = None
    ) -> frozenset:
        """
        Unpack a frozenset from file.

        :param data_item:   The data item to unpack.
        :param file_format: The file format to use for reading the frozenset. Default is None - will be read by the file
                            extension.

        :return: The unpacked frozenset.
        """
        return frozenset(
            super().unpack_file(data_item=data_item, file_format=file_format)
        )


class BytesPackager(ListPackager):
    """
    ``builtins.bytes`` packager.
    """

    PACKABLE_OBJECT_TYPE = bytes

    def pack_result(self, obj: bytes, key: str) -> dict:
        """
        Pack bytes as a result.

        :param obj: The bytearray to pack and log.
        :param key: The result's key.

        :return: The result dictionary.
        """
        return {key: obj}

    def pack_file(
        self, obj: bytes, key: str, file_format: str = DEFAULT_STRUCT_FILE_FORMAT
    ) -> tuple[Artifact, dict]:
        """
        Pack a bytes as a file by the given format.

        :param obj:         The bytes to pack.
        :param key:         The key to use for the artifact.
        :param file_format: The file format to save as. Default is json.

        :return: The packed artifact and instructions.
        """
        return super().pack_file(obj=list(obj), key=key, file_format=file_format)

    def unpack_file(
        self, data_item: DataItem, file_format: Optional[str] = None
    ) -> bytes:
        """
        Unpack a bytes from file.

        :param data_item:   The data item to unpack.
        :param file_format: The file format to use for reading the bytes. Default is None - will be read by the file
                            extension.

        :return: The unpacked bytes.
        """
        return bytes(super().unpack_file(data_item=data_item, file_format=file_format))


class BytearrayPackager(BytesPackager):
    """
    ``builtins.bytearray`` packager.
    """

    PACKABLE_OBJECT_TYPE = bytearray

    def pack_result(self, obj: bytearray, key: str) -> dict:
        """
        Pack a bytearray as a result.

        :param obj: The bytearray to pack and log.
        :param key: The result's key.

        :return: The result dictionary.
        """
        return {key: bytes(obj)}

    def pack_file(
        self, obj: bytearray, key: str, file_format: str = DEFAULT_STRUCT_FILE_FORMAT
    ) -> tuple[Artifact, dict]:
        """
        Pack a bytearray as a file by the given format.

        :param obj:         The bytearray to pack.
        :param key:         The key to use for the artifact.
        :param file_format: The file format to save as. Default is json.

        :return: The packed artifact and instructions.
        """
        return super().pack_file(obj=bytes(obj), key=key, file_format=file_format)

    def unpack_file(
        self, data_item: DataItem, file_format: Optional[str] = None
    ) -> bytearray:
        """
        Unpack a bytearray from file.

        :param data_item:   The data item to unpack.
        :param file_format: The file format to use for reading the bytearray. Default is None - will be read by the file
                            extension.

        :return: The unpacked bytearray.
        """
        return bytearray(
            super().unpack_file(data_item=data_item, file_format=file_format)
        )


# ----------------------------------------------------------------------------------------------------------------------
# pathlib packagers:
# ----------------------------------------------------------------------------------------------------------------------


class PathPackager(StrPackager):
    """
    ``pathlib.Path`` packager. It is also used for all `Path` inheriting pathlib objects (`PosixPath` and
    `WindowsPath`).
    """

    PACKABLE_OBJECT_TYPE = pathlib.Path
    PACK_SUBCLASSES = True
    DEFAULT_PACKING_ARTIFACT_TYPE = "path"

    def pack_result(self, obj: pathlib.Path, key: str) -> dict:
        """
        Pack the `Path` as a string result.

        :param obj: The `Path` to pack.
        :param key: The key to use in the results dictionary.

        :return: The packed result.
        """
        return super().pack_result(obj=str(obj), key=key)

    def pack_path(
        self, obj: pathlib.Path, key: str, archive_format: str = DEFAULT_ARCHIVE_FORMAT
    ) -> tuple[Artifact, dict]:
        """
        Pack a `Path` value (pack the file or directory in that path).

        :param obj:            The `Path` to pack.
        :param key:            The key to use for the artifact.
        :param archive_format: The archive format to use in case the path is of a directory. Default is zip.

        :return: The packed artifact and instructions.
        """
        return super().pack_path(obj=str(obj), key=key, archive_format=archive_format)

    def unpack_path(
        self,
        data_item: DataItem,
        is_directory: bool = False,
        archive_format: Optional[str] = None,
    ) -> pathlib.Path:
        """
        Unpack a data item representing a `Path`. If the path is of a file, the file is downloaded to a local
        temporary directory and its path is returned. If the path is of a directory, the archive is extracted and the
        directory path extracted is returned.

        :param data_item:      The data item to unpack.
        :param is_directory:   Whether the path should be treated as a file or a directory. Files (even archives like
                               zip) won't be extracted.
        :param archive_format: The archive format to use in case the path is of a directory. Default is None - will be
                               read by the archive file extension.

        :return: The unpacked `Path`.
        """
        return pathlib.Path(
            super().unpack_path(
                data_item=data_item,
                is_directory=is_directory,
                archive_format=archive_format,
            )
        )


# ----------------------------------------------------------------------------------------------------------------------
# TODO: collection packagers:
# ----------------------------------------------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------------------------------------------
# TODO: datetime packagers:
# ----------------------------------------------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------------------------------------------
# TODO: enum packagers:
# ----------------------------------------------------------------------------------------------------------------------



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Configuration system.

Configuration can be in either a configuration file specified by
MLRUN_CONFIG_FILE environment variable or by environment variables.

Environment variables are in the format "MLRUN_HTTPDB__PORT=8080". This will be
mapped to config.httpdb.port. Values should be in JSON format.
"""

import base64
import binascii
import copy
import json
import os
import typing
import warnings
from collections.abc import Mapping
from datetime import timedelta
from os.path import expanduser
from threading import Lock

import dotenv
import semver
import urllib3.exceptions
import yaml

import mlrun.common.constants
import mlrun.common.schemas
import mlrun.errors

env_prefix = "MLRUN_"
env_file_key = f"{env_prefix}CONFIG_FILE"
_load_lock = Lock()
_none_type = type(None)
default_env_file = os.getenv("MLRUN_DEFAULT_ENV_FILE", "~/.mlrun.env")


default_config = {
    "namespace": "",  # default kubernetes namespace
    "kubernetes": {
        "kubeconfig_path": "",  # local path to kubeconfig file (for development purposes),
        # empty by default as the API already running inside k8s cluster
        "pagination": {
            # pagination config for interacting with k8s API
            "list_pods_limit": 200,
            "list_crd_objects_limit": 200,
        },
    },
    "dbpath": "",  # db/api url
    # url to nuclio dashboard api (can be with user & token, e.g. https://username:password@dashboard-url.com)
    "nuclio_dashboard_url": "",
    "nuclio_version": "",
    "default_nuclio_runtime": "python:3.11",
    "nest_asyncio_enabled": "",  # enable import of nest_asyncio for corner cases with old jupyter, set "1"
    "ui_url": "",  # remote/external mlrun UI url (for hyperlinks) (This is deprecated in favor of the ui block)
    "remote_host": "",
    "api_base_version": "v1",
    "version": "",  # will be set to current version
    "images_tag": "",  # tag to use with mlrun images e.g. mlrun/mlrun (defaults to version)
    # registry to use with mlrun images that start with "mlrun/" e.g. quay.io/ (defaults to empty, for dockerhub)
    "images_registry": "",
    # registry to use with non-mlrun images (don't start with "mlrun/") specified in 'images_to_enrich_registry'
    # defaults to empty, for dockerhub
    "vendor_images_registry": "",
    # comma separated list of images that are in the specified images_registry, and therefore will be enriched with this
    # registry when used. default to mlrun/* which means any image which is of the mlrun repository (mlrun/mlrun,
    # mlrun/mlrun-kfp, etc...)
    "images_to_enrich_registry": "^mlrun/*,^python:3.(9|11)$",
    "kfp_url": "",
    "kfp_ttl": "14400",  # KFP ttl in sec, after that completed PODs will be deleted
    "kfp_image": "mlrun/mlrun-kfp",  # image to use for KFP runner
    "dask_kfp_image": "mlrun/mlrun",  # image to use for dask KFP runner
    "igz_version": "",  # the version of the iguazio system the API is running on
    "iguazio_api_url": "",  # the url to iguazio api
    "spark_app_image": "",  # image to use for spark operator app runtime
    "spark_app_image_tag": "",  # image tag to use for spark operator app runtime
    "spark_history_server_path": "",  # spark logs directory for spark history server
    "spark_operator_version": "spark-3",  # the version of the spark operator in use
    "package_path": "mlrun",  # mlrun pip package
    "default_base_image": "mlrun/mlrun",  # default base image when doing .deploy()
    # template for project default image name. Parameter {name} will be replaced with project name
    "default_project_image_name": ".mlrun-project-image-{name}",
    "active_project": "",  # active project name
    "default_archive": "",  # default remote archive URL (for build tar.gz)
    "mpijob_crd_version": "",  # mpijob crd version (e.g: "v1alpha1". must be in: mlrun.runtime.MPIJobCRDVersions)
    "ipython_widget": True,
    "log_level": "INFO",
    # log formatter (options: human | human_extended | json)
    "log_formatter": "human",
    # custom logger format, workes only with log_formatter: custom
    # Note that your custom format must include those 4 fields - timestamp, level, message and more
    "log_format_override": None,
    "submit_timeout": "280",  # timeout when submitting a new k8s resource
    # runtimes cleanup interval in seconds
    "runtimes_cleanup_interval": "300",
    # disabled by default due to an internal bug in serving functions
    # relying on a background task to hold the status for its model endpoints
    # TODO: need to refine what/when we can delete the background tasks
    # e.g: use labels or naming convention.
    "background_task_cleanup_interval": "0",
    "background_task_max_age": "21600",  # 6 hours in seconds
    "monitoring": {
        "runs": {
            # runs monitoring interval in seconds
            "interval": "30",
            # runs monitoring debouncing interval in seconds for run with non-terminal state without corresponding
            # k8s resource by default the interval will be - (monitoring.runs.interval * 2 ), if set will override the
            # default
            "missing_runtime_resources_debouncing_interval": None,
            # max number of parallel abort run jobs in runs monitoring
            "concurrent_abort_stale_runs_workers": 10,
            "list_runs_time_period_in_days": 7,  # days
            "retry": {
                # periodic job for triggering retries interval in seconds
                "interval": "30",
                # runs limit to fetch for retrying
                "fetch_runs_limit": 1000,
                # minutes until a run is considered stale and will be aborted
                "staleness_threshold": 60 * 24 * 3,
            },
        },
        "projects": {
            "summaries": {
                "cache_interval": "30",
            },
        },
    },
    "crud": {
        "runs": {
            # deleting runs is a heavy operation that includes deleting runtime resources, therefore we do it in chunks
            "batch_delete_runs_chunk_size": 10,
        },
        "resources": {
            "delete_crd_resources_timeout": "5 minutes",
        },
    },
    "object_retentions": {
        "alert_activations": 14 * 7,  # days
    },
    # A safety margin to account for delays
    # This ensures that extra partitions are available beyond the specified retention period
    "partitions_buffer_multiplier": 3,
    # the grace period (in seconds) that will be given to runtime resources (after they're in terminal state)
    # before deleting them (4 hours)
    "runtime_resources_deletion_grace_period": "14400",
    "scrape_metrics": True,
    # sets the background color that is used in printed tables in jupyter
    "background_color": "#4EC64B",
    "artifact_path": "",  # default artifacts path/url
    # Add {{workflow.uid}} to artifact_path unless user specified a path manually
    "enrich_artifact_path_with_workflow_id": True,
    "artifacts": {
        "calculate_hash": True,
        # None is handled as False, reason we set None instead of False is that if the server have set the value to
        # some value while the client didn't change it, the server value will be applied.
        # But if both the server and the client set some value, we want the client to take precedence over the server.
        # By setting the default to None we are able to differentiate between the two cases.
        "generate_target_path_from_artifact_hash": None,
        # migration from artifacts to artifacts_v2 is done in batches, and requires a state file to keep track of the
        # migration progress.
        "artifact_migration_batch_size": 200,
        "artifact_migration_v9_batch_size": 30000,
        "artifact_migration_state_file_path": "./db/_artifact_migration_state.json",
        "datasets": {
            "max_preview_columns": 100,
        },
        "limits": {
            "max_chunk_size": 1024 * 1024 * 1,  # 1MB
            "max_preview_size": 1024 * 1024 * 10,  # 10MB
            "max_download_size": 1024 * 1024 * 100,  # 100MB
            "max_deletions": 200,
        },
    },
    # FIXME: Adding these defaults here so we won't need to patch the "installing component" (provazio-controller) to
    #  configure this values on field systems, for newer system this will be configured correctly
    "v3io_api": "http://v3io-webapi:8081",
    "redis": {
        "url": "",
        "type": "standalone",  # deprecated.
    },
    "sql": {
        "url": "",
    },
    "v3io_framesd": "http://framesd:8080",
    "model_providers": {
        "openai_default_model": "gpt-4o",
        "huggingface_default_model": "microsoft/Phi-3-mini-4k-instruct",
    },
    # default node selector to be applied to all functions - json string base64 encoded format
    "default_function_node_selector": "e30=",
    # default priority class to be applied to functions running on k8s cluster
    "default_function_priority_class_name": "",
    # valid options for priority classes - separated by a comma
    "valid_function_priority_class_names": "",
    # default path prefix for demo data and models
    "default_samples_path": "https://s3.wasabisys.com/iguazio/",
    # default path for tensorboard logs
    "default_tensorboard_logs_path": "/User/.tensorboard/{{project}}",
    # ";" separated list of notebook cell tag names to ignore e.g. "ignore-this;ignore-that"
    "ignored_notebook_tags": "",
    # when set it will force the local=True in run_function(), set to "auto" will run local if there is no k8s
    "force_run_local": "auto",
    # when set (True or non empty str) it will force the mock=True in deploy_function(),
    # set to "auto" will use mock of Nuclio if not detected (no nuclio_version)
    "mock_nuclio_deployment": "",
    # Configurations for `mlrun.track` - tracking runs and experiments from 3rd party vendors like MLFlow
    # by running them as a MLRun function, capturing their logs, results and artifacts to mlrun.
    "external_platform_tracking": {
        # General enabler for the entire tracking mechanism (all tracking services):
        "enabled": False,
        # Specific enablement and other configurations for the supported trackers:
        "mlflow": {
            # Enabler of MLFlow tracking:
            "enabled": True,
            # Whether to match the experiment name to the runtime name (sets mlflow experiment name to mlrun
            # context name):
            "match_experiment_to_runtime": False,
            # Whether to determine the mlflow run id before tracking starts, by doing so we can be positive that we
            # are tracking the correct run, this is useful especially for when we run number of runs simultaneously
            # in the same experiment. the default is set to false because in the process a mlflow run is created in
            # advance, and we want to avoid creating unnecessary runs.
            "control_run": False,
        },
    },
    "background_tasks": {
        # enabled / disabled
        "timeout_mode": "enabled",
        "function_deletion_batch_size": 10,
        # timeout in seconds to wait for background task to be updated / finished by the worker responsible for the task
        "default_timeouts": {
            "operations": {
                "migrations": "3600",
                "load_project": "60",
                "run_abortion": "600",
                "abort_grace_period": "10",
                "delete_project": "900",
                "delete_function": "900",
                "model_endpoint_creation": "600",
                "model_endpoint_tsdb_leftovers": "900",
                "terminate_pipeline": "300",
            },
            "runtimes": {
                "dask": "600",
                # cluster start might take some time in case k8s needs to spin up new nodes
                "dask_cluster_start": "600",
            },
            "push_notifications": "60",
        },
    },
    "function": {
        "spec": {
            "image_pull_secret": {"default": None},
            "security_context": {
                # default security context to be applied to all functions - json string base64 encoded format
                # in camelCase format: {"runAsUser": 1000, "runAsGroup": 3000}
                "default": "e30=",  # encoded empty dict
                # see mlrun.common.schemas.function.SecurityContextEnrichmentModes for available options
                "enrichment_mode": "disabled",
                # default 65534 (nogroup), set to -1 to use the user unix id or
                # function.spec.security_context.pipelines.kfp_pod_user_unix_id for kfp pods
                "enrichment_group_id": 65534,
                "pipelines": {
                    # sets the user id to be used for kfp pods when enrichment mode is not disabled
                    "kfp_pod_user_unix_id": 5,
                },
            },
            "service_account": {"default": None},
            "state_thresholds": {
                "default": {
                    "pending_scheduled": "1h",
                    "pending_not_scheduled": "-1",  # infinite
                    "image_pull_backoff": "1h",
                    "executing": "24h",
                }
            },
            "retry": {
                "backoff": {
                    "default_base_delay": "30s",
                    "min_base_delay": "30s",
                },
            },
            # When the module is reloaded, the maximum depth recursion configuration for the recursive reload
            # function is used to prevent infinite loop
            "reload_max_recursion_depth": 100,
            "source_code_max_bytes": 10000,
        },
        "databricks": {
            "artifact_directory_path": "/mlrun_databricks_runtime/artifacts_dictionaries"
        },
        "application": {
            "default_sidecar_internal_port": 8050,
            "default_authentication_mode": mlrun.common.schemas.APIGatewayAuthenticationMode.none,
            "default_worker_number": 10000,
        },
    },
    # TODO: function defaults should be moved to the function spec config above
    "function_defaults": {
        "image_by_kind": {
            "job": "mlrun/mlrun",
            "serving": "mlrun/mlrun",
            "nuclio": "mlrun/mlrun",
            "remote": "mlrun/mlrun",
            "dask": "mlrun/mlrun",
            "mpijob": "mlrun/mlrun",
            "application": "python",
        },
        # see enrich_function_preemption_spec for more info,
        # and mlrun.common.schemas.function.PreemptionModes for available options
        "preemption_mode": "prevent",
    },
    "httpdb": {
        "clusterization": {
            # one of chief/worker
            "role": "chief",
            "chief": {
                # when url is specified, it takes precedence over service and port
                "url": "",
                "service": "mlrun-api-chief",
                "port": 8080,
                "feature_gates": {
                    "scheduler": "enabled",
                    "project_sync": "enabled",
                    "cleanup": "enabled",
                    "runs_monitoring": "enabled",
                    "pagination_cache": "enabled",
                    "project_summaries": "enabled",
                    "start_logs": "enabled",
                    "stop_logs": "enabled",
                    "retry_jobs": "enabled",
                },
            },
            "worker": {
                "sync_with_chief": {
                    # enabled / disabled
                    "mode": "enabled",
                    "interval": 15,  # seconds
                },
                "request_timeout": 45,  # seconds
            },
            # see server.py.services.api.utils.helpers.ensure_running_on_chief
            "ensure_function_running_on_chief_mode": "enabled",
        },
        "port": 8080,
        "dirpath": expanduser("~/.mlrun/db"),
        # in production envs we recommend to use a real db (e.g. mysql)
        "dsn": "sqlite:///db/mlrun.db?check_same_thread=false",
        "debug": False,
        "user": "",
        "password": "",
        "token": "",
        "logs_path": "./db/logs",
        # when set, these will replace references to the data_volume with the real_path
        "data_volume": "",
        "real_path": "",
        # comma delimited prefixes of paths allowed through the /files API (v3io & the real_path are always allowed).
        # These paths must be schemas (cannot be used for local files). For example "s3://mybucket,gcs://"
        "allowed_file_paths": "s3://,gcs://,gs://,az://,dbfs://,ds://",
        "db_type": "sqldb",
        "max_workers": 64,
        # See mlrun.common.schemas.APIStates for options
        "state": "online",
        "retry_api_call_on_exception": "enabled",
        "http_connection_timeout_keep_alive": 11,
        # http client used by httpdb
        "http": {
            # when True, the client will verify the server's TLS
            # set to False for backwards compatibility.
            "verify": True,
        },
        "db": {
            "commit_retry_timeout": 30,
            "commit_retry_interval": 3,
            "conflict_retry_timeout": 15,
            "conflict_retry_interval": None,
            # Whether to perform data migrations on initialization. enabled or disabled
            "data_migrations_mode": "enabled",
            # Whether to perform database migration from sqlite to mysql on initialization
            "database_migration_mode": "enabled",
            "backup": {
                # Whether to use db backups on initialization
                "mode": "enabled",
                "file_format": "db_backup_%Y%m%d%H%M.db",
                "use_rotation": True,
                "rotation_limit": 3,
                # default is 16MB, max 1G, for more info https://dev.mysql.com/doc/refman/8.0/en/packet-too-large.html
                "max_allowed_packet": 64000000,  # 64MB
            },
            # tests connections for liveness upon each checkout
            "connections_pool_pre_ping": True,
            # this setting causes the pool to recycle connections after the given number of seconds has passed
            "connections_pool_recycle": 60 * 60,
            # None defaults to httpdb.max_workers
            "connections_pool_size": None,
            "connections_pool_max_overflow": None,
            # below is a db-specific configuration
            "mysql": {
                # comma separated mysql modes (globally) to set on runtime
                # optional values (as per https://dev.mysql.com/doc/refman/8.0/en/sql-mode.html#sql-mode-full):
                #
                # if set to "nil" or "none", nothing would be set
                "modes": (
                    "STRICT_TRANS_TABLES,NO_ZERO_IN_DATE,NO_ZERO_DATE,ERROR_FOR_DIVISION_BY_ZERO,NO_ENGINE_SUBSTITUTION"
                )
            },
        },
        "jobs": {
            # whether to allow to run local runtimes in the API - configurable to allow the scheduler testing to work
            "allow_local_run": False,
        },
        "authentication": {
            "mode": "none",  # one of none, basic, bearer, iguazio
            "basic": {"username": "", "password": ""},
            "bearer": {"token": ""},
            "iguazio": {
                "session_verification_endpoint": "data_sessions/verifications/app_service",
            },
        },
        "nuclio": {
            # One of ClusterIP | NodePort
            "default_service_type": "NodePort",
            # The following modes apply when user did not configure an ingress
            #
            #   name        |  description
            #  ---------------------------------------------------------------------
            #   never       |  never enrich with an ingress
            #   always      |  always enrich with an ingress, regardless the service type
            #   onClusterIP |  enrich with an ingress only when `mlrun.config.httpdb.nuclio.default_service_type`
            #                  is set to ClusterIP
            #  ---------------------------------------------------------------------
            # Note: adding a mode requires special handling on
            # - mlrun.common.runtimes.constants.NuclioIngressAddTemplatedIngressModes
            # - mlrun.runtimes.nuclio.function.enrich_function_with_ingress
            "add_templated_ingress_host_mode": "never",
            "explicit_ack": "enabled",
            # size of serving spec to move to config maps
            "serving_spec_env_cutoff": 0,
        },
        "logs": {
            "decode": {
                # Replace with a replacement marker. Uses � (U+FFFD, the official REPLACEMENT CHARACTER).
                # see https://docs.python.org/3/library/codecs.html#error-handlers for more info and options
                "errors": "replace",
            },
            "pipelines": {
                # pull state mode was introduced to have a way to pull the state of a run which was spawned by a
                # pipeline step instead of pulling the state by getting the run logs
                "pull_state": {
                    # enabled - pull state of a run every "pull_state_interval" seconds and pull logs every
                    # "pull_logs_interval" seconds
                    # disabled - pull logs every "pull_logs_default_interval" seconds
                    "mode": "disabled",
                    # those params are used when mode is enabled
                    "pull_logs_interval": 30,  # seconds
                    "pull_state_interval": 5,  # seconds
                },
            },
            "nuclio": {
                # setting interval to a higher interval than regular jobs / build, because pulling the retrieved logs
                # from nuclio for the deploy status doesn't include the actual live "builder" container logs, but
                # rather a high level status
                "pull_deploy_status_default_interval": 10  # seconds
            },
            # this is the default interval period for pulling logs, if not specified different timeout interval
            "pull_logs_default_interval": 3,  # seconds
            "pull_logs_backoff_no_logs_default_interval": 10,  # seconds
            "pull_logs_default_size_limit": 1024 * 1024,  # 1 MB
        },
        "authorization": {
            "mode": "none",  # one of none, opa
            "opa": {
                "address": "",
                "request_timeout": 10,
                "permission_query_path": "",
                "permission_filter_path": "",
                "log_level": 0,
            },
        },
        "scheduling": {
            # the minimum interval that will be allowed between two scheduled jobs - e.g. a job wouldn't be
            # allowed to be scheduled to run more than 2 times in X. Can't be less than 1 minute, "0" to disable
            "min_allowed_interval": "10 minutes",
            "default_concurrency_limit": 1,
            # Firing our jobs include things like creating pods which might not be instant, therefore in the case of
            # multiple schedules scheduled to the same time, there might be delays, the default of the scheduler for
            # misfire_grace_time is 1 second, we do not want jobs not being scheduled because of the delays so setting
            # it to None. the default for coalesce it True just adding it here to be explicit
            "scheduler_config": '{"job_defaults": {"misfire_grace_time": null, "coalesce": true}}',
        },
        "projects": {
            "leader": "mlrun",
            "retry_leader_request_on_exception": "enabled",
            "followers": "",
            # This is used as the interval for the sync loop both when mlrun is leader and follower
            "periodic_sync_interval": "1 minute",
            "project_owners_cache_ttl": "30 seconds",
            # access key to be used when the leader is iguazio and polling is done from it
            "iguazio_access_key": "",
            "iguazio_list_projects_default_page_size": 200,
            "iguazio_client_job_cache_ttl": "20 minutes",
            "nuclio_project_deletion_verification_timeout": "300 seconds",
            "nuclio_project_deletion_verification_interval": "5 seconds",
            "summaries": {
                # Number of days back to include when calculating the project pipeline summary.
                "list_pipelines_time_period_in_days": 7,
            },
            "resource_deletion_batch_size": 10000,
        },
        # The API needs to know what is its k8s svc url so it could enrich it in the jobs it creates
        "api_url": "",
        "builder": {
            # setting the docker registry to be used for built images, can include the repository as well, e.g.
            # index.docker.io/<username>, if not included repository will default to mlrun
            "docker_registry": "",
            # dockerconfigjson type secret to attach to kaniko pod.
            # For amazon ECR, the secret is expected to provide AWS credentials. Leave empty to use EC2 IAM policy.
            # https://github.com/GoogleContainerTools/kaniko#pushing-to-amazon-ecr
            "docker_registry_secret": "",
            # whether to allow the docker registry we're pulling from to be insecure. "enabled", "disabled" or "auto"
            # which will resolve by the existence of secret
            "insecure_pull_registry_mode": "auto",
            # whether to allow the docker registry we're pushing to, to be insecure. "enabled", "disabled" or "auto"
            # which will resolve by the existence of secret
            "insecure_push_registry_mode": "auto",
            # the requirement specifier used by the builder when installing mlrun in images when it runs
            # pip install <requirement_specifier>, e.g. mlrun==0.5.4, mlrun~=0.5,
            # git+https://github.com/mlrun/mlrun@development. by default uses the version
            "mlrun_version_specifier": "",
            "kaniko_image": "gcr.io/kaniko-project/executor:v1.23.2",  # kaniko builder image
            "kaniko_init_container_image": "alpine:3.20",
            # image for kaniko init container when docker registry is ECR
            "kaniko_aws_cli_image": "amazon/aws-cli:2.17.16",
            # kaniko sometimes fails to get filesystem from image, this is a workaround to retry the process
            # a known issue in Kaniko - https://github.com/GoogleContainerTools/kaniko/issues/1717
            "kaniko_image_fs_extraction_retries": "3",
            # kaniko sometimes fails to push image to registry due to network issues
            "kaniko_image_push_retry": "3",
            # additional docker build args in json encoded base64 format
            "build_args": "",
            "pip_ca_secret_name": "",
            "pip_ca_secret_key": "",
            "pip_ca_path": "/etc/ssl/certs/mlrun/pip-ca-certificates.crt",
            # template for the prefix that the function target image will be enforced to have (as long as it's targeted
            # to be in the configured registry). Supported template values are: {project} {name}
            "function_target_image_name_prefix_template": "func-{project}-{name}",
            "pip_version": "~=23.0",
        },
        "v3io_api": "",
        "v3io_framesd": "",
        # If running from sdk and MLRUN_DBPATH is not set, the db will fallback to a nop db which will not perform any
        # run db operations.
        "nop_db": {
            # if set to true, will raise an error for trying to use run db functionality
            # if set to false, will use a nop db which will not preform any run db operations
            "raise_error": False,
            # if set to true, will log a warning for trying to use run db functionality while in nop db mode
            "verbose": True,
        },
        "pagination": {
            "default_page_size": 200,
            "page_limit": 1000000,
            "page_size_limit": 1000000,
            "pagination_cache": {
                "interval": 60,
                "ttl": 3600,
                "max_size": 10000,
            },
        },
    },
    "model_endpoint_monitoring": {
        # Scaling Rule
        # The fundamental scaling rule to maintain is: Shards/Partitions = Replicas * Workers
        # In other words, the number of shards (V3IO) or partitions (Kafka) must be equal to the
        # total number of worker processes across all pods.
        "serving_stream": {
            "v3io": {
                "shard_count": 2,
                "retention_period_hours": 24,
                "num_workers": 1,
                "min_replicas": 2,
                "max_replicas": 2,
            },
            "kafka": {
                "partition_count": 8,
                "replication_factor": 1,
                "num_workers": 2,
                "min_replicas": 1,
                "max_replicas": 4,
            },
        },
        "application_stream_args": {
            "v3io": {
                "shard_count": 4,
                "retention_period_hours": 24,
                "num_workers": 4,
                "min_replicas": 1,
                "max_replicas": 1,
            },
            "kafka": {
                "partition_count": 4,
                "replication_factor": 1,
                "num_workers": 4,
                "min_replicas": 1,
                "max_replicas": 1,
            },
        },
        "writer_stream_args": {
            "v3io": {
                "shard_count": 4,
                "retention_period_hours": 24,
                "num_workers": 4,
                "min_replicas": 1,
                "max_replicas": 1,
            },
            "kafka": {
                "partition_count": 4,
                # TODO: add retention period configuration
                "replication_factor": 1,
                "num_workers": 4,
                "min_replicas": 1,
                "max_replicas": 1,
            },
        },
        "controller_stream_args": {
            "v3io": {
                "shard_count": 10,
                "retention_period_hours": 24,
                "num_workers": 10,
                "min_replicas": 1,
                "max_replicas": 1,
            },
            "kafka": {
                "partition_count": 10,
                "replication_factor": 1,
                "num_workers": 10,
                "min_replicas": 1,
                "max_replicas": 1,
            },
        },
        "writer_graph": {
            "max_events": 1000,
            "flush_after_seconds": 30,
            "writer_version": "v1",  # v1 is the sync version while v2 is async
            "parquet_batching_max_events": 10,
            "parquet_batching_timeout_secs": 30,
        },
        # Store prefixes are used to handle model monitoring storing policies based on project and kind, such as events,
        # stream, and endpoints.
        "store_prefixes": {
            "default": "v3io:///users/pipelines/{project}/model-endpoints/{kind}",
            "user_space": "v3io:///projects/{project}/model-endpoints/{kind}",
            "monitoring_application": "v3io:///users/pipelines/{project}/monitoring-apps/",
        },
        # Offline storage path can be either relative or a full path. This path is used for general offline data
        # storage such as the parquet file which is generated from the monitoring stream function for the drift analysis
        "offline_storage_path": "model-endpoints/{kind}",
        "parquet_batching_max_events": 10_000,
        "parquet_batching_timeout_secs": timedelta(minutes=1).total_seconds(),
        "model_endpoint_creation_check_period": 15,
    },
    "secret_stores": {
        # Use only in testing scenarios (such as integration tests) to avoid using k8s for secrets (will use in-memory
        # "secrets")
        "test_mode_mock_secrets": False,
        "vault": {
            # URLs to access Vault. For example, in a local env (Minikube on Mac) these would be:
            # http://docker.for.mac.localhost:8200
            "url": "",
            "remote_url": "",
            "role": "",
            "token_path": "~/.mlrun/vault",
            "project_service_account_name": "mlrun-vault-{project}",
            "token_ttl": 180000,
            # This config is for debug/testing purposes only!
            "user_token": "",
        },
        "azure_vault": {
            "url": "https://{name}.vault.azure.net",
            "default_secret_name": None,
            "secret_path": "~/.mlrun/azure_vault",
        },
        "kubernetes": {
            # When this is True (the default), all project secrets will be automatically added to each job,
            # unless user asks for a specific list of secrets.
            "auto_add_project_secrets": True,
            "project_secret_name": "mlrun-project-secrets-{project}",
            "auth_secret_name": "mlrun-auth-secrets.{hashed_access_key}",
            "env_variable_prefix": "",
            "global_function_env_secret_name": None,
        },
    },
    "feature_store": {
        "data_prefixes": {
            "default": "v3io:///projects/{project}/FeatureStore/{name}/{kind}",
            "nosql": "v3io:///projects/{project}/FeatureStore/{name}/nosql",
            # "authority" is optional and generalizes [userinfo "@"] host [":" port]
            "redisnosql": "redis://{authority}/projects/{project}/FeatureStore/{name}/nosql",
            "dsnosql": "ds://{ds_profile_name}/projects/{project}/FeatureStore/{name}/{kind}",
        },
        "default_targets": "parquet,nosql",
        "default_job_image": "mlrun/mlrun",
        "flush_interval": None,
    },
    "ui": {
        "projects_prefix": "projects",  # The UI link prefix for projects
        "url": "",  # remote/external mlrun UI url (for hyperlinks)
    },
    "hub": {
        "k8s_secrets_project_name": "-hub-secrets",
        "catalog_filename": "catalog.json",
        "default_source": {
            # Set false to avoid creating a global source (for example in a dark site)
            "create": True,
            "name": "default",
            "description": "MLRun global function hub",
            "url": "https://mlrun.github.io/marketplace",
            "channel": "master",
        },
    },
    "storage": {
        # What type of auto-mount to use for functions. One of: none, auto, v3io_credentials, v3io_fuse, pvc, s3, env.
        # Default is auto - which is v3io_credentials when running on Iguazio. If not Iguazio: pvc if the
        # MLRUN_PVC_MOUNT env is configured or auto_mount_params contain "pvc_name". Otherwise will do nothing (none).
        "auto_mount_type": "auto",
        # Extra parameters to pass to the mount call (will be passed as kwargs). Parameters can be either:
        # 1. A string of comma-separated parameters, using this format: "param1=value1,param2=value2"
        # 2. A base-64 encoded json dictionary containing the list of parameters
        "auto_mount_params": "",
        # map file data items starting with virtual path to the real path, used when consumers have different mounts
        # e.g. Windows client (on host) and Linux container (Jupyter, Nuclio..) need to access the same files/artifacts
        # need to map container path to host windows paths, e.g. "\data::c:\\mlrun_data" ("::" used as splitter)
        "item_to_real_path": "",
    },
    "default_function_pod_resources": {
        "requests": {"cpu": None, "memory": None, "gpu": None},
        "limits": {"cpu": None, "memory": None, "gpu": None},
    },
    "default_spark_resources": {
        "driver": {
            "requests": {"cpu": "1", "memory": "2g"},
            "limits": {"cpu": "2", "memory": "2g"},
        },
        "executor": {
            "requests": {"cpu": "1", "memory": "5g"},
            "limits": {"cpu": "2", "memory": "5g"},
        },
    },
    # preemptible node selector and tolerations to be added when running on spot nodes
    "preemptible_nodes": {
        # encoded empty dict
        "node_selector": "e30=",
        # encoded empty list
        "tolerations": "W10=",
    },
    "http_retry_defaults": {
        "max_retries": 3,
        "backoff_factor": 1,
        "status_codes": [500, 502, 503, 504],
    },
    "ce": {
        # ce mode can be one of: "", lite, full
        "mode": "",
        # not possible to call this "version" because the Config class has a "version" property
        # which returns the version from the version.json file
        "release": "",
    },
    "debug": {
        "expose_internal_api_endpoints": False,
    },
    "workflows": {
        "default_workflow_runner_name": "workflow-runner-{}",
        "concurrent_delete_worker_count": 20,
        # Default timeout seconds for retrieving workflow id after execution
        # Remote workflow timeout is the maximum between remote and the inner engine timeout
        "timeouts": {"local": 120, "kfp": 60, "remote": 60 * 5},
    },
    "log_collector": {
        "address": "localhost:8282",
        # log collection mode can be one of: "sidecar", "legacy", "best-effort"
        # "sidecar" - use the sidecar to collect logs
        # "legacy" - use the legacy log collection method (logs are collected straight from the pod)
        # "best-effort" - use the sidecar, but if for some reason it's not available use the legacy method
        # note that this mode also effects the log querying method as well, meaning if the mode is "best-effort"
        # the log query will try to use the sidecar first and if it's not available it will use the legacy method
        # TODO: once this is changed to "sidecar" by default, also change in common_fixtures.py
        "mode": "legacy",
        # interval for collecting and sending runs which require their logs to be collected
        "periodic_start_log_interval": 10,
        "failed_runs_grace_period": 3600,
        "verbose": True,
        # the number of workers which will be used to trigger the start log collection
        "concurrent_start_logs_workers": 50,
        # the number of runs for which to start logs on api startup
        "start_logs_startup_run_limit": 150,
        # the time in hours in which to start log collection from.
        # after upgrade, we might have runs which completed in the mean time or still in non-terminal state and
        # we want to collect their logs in the new log collection method (sidecar)
        # default is 4 hours = 4*60*60 = 14400 seconds
        "api_downtime_grace_period": 14400,
        "get_logs": {
            # the number of retries to get logs from the log collector
            "max_retries": 3,
        },
        # interval for stopping log collection for runs which are in a terminal state
        "stop_logs_interval": 3600,
    },
    # Configurations for the `mlrun.package` sub-package involving packagers - logging returned outputs and parsing
    # inputs data items:
    "packagers": {
        # Whether to enable packagers. True will wrap each run in the `mlrun.package.handler` decorator to log and parse
        # using packagers.
        "enabled": True,
        # Whether to treat returned tuples from functions as a tuple and not as multiple returned items. If True, all
        # returned values will be packaged together as the tuple they are returned in. Default is False to enable
        # logging multiple returned items.
        "pack_tuples": False,
        # In multi-workers run, only the logging worker will pack the outputs and log the results and artifacts.
        # Otherwise, the workers will log the results and artifacts using the same keys, overriding them. It is common
        # that only the main worker (usualy rank 0) will log, so this is the default value.
        "logging_worker": 0,
        # TODO: Consider adding support for logging from all workers (ignoring the `logging_worker`) and add the worker
        #       number to the artifact / result key (like "<key>-rank<#>". Results can have reduce operation in the
        #       log hint to average / min / max them across all the workers (default operation should be average).
    },
    # Events are currently (and only) used to audit changes and record access to MLRun entities (such as secrets)
    "events": {
        # supported modes "enabled", "disabled".
        # "enabled" - events are emitted.
        # "disabled" - a nop client is used (aka doing nothing).
        "mode": "disabled",
        "verbose": False,
        # used for igz client when emitting events
        "access_key": "",
    },
    "grafana_url": "",
    "alerts": {
        # supported modes: "enabled", "disabled".
        "mode": "enabled",
        # maximum number of alerts we allow to be configured.
        # user will get an error when exceeding this
        "max_allowed": 20000,
        # maximum allowed value for count in criteria field inside AlertConfig
        "max_criteria_count": 100,
        # interval for periodic events generation job
        "events_generation_interval": 30,  # seconds
        # number of alerts to delete in each chunk
        "chunk_size_during_project_deletion": 100,
        # maximum allowed alert config cache size in alert's CRUD
        # for the best performance, it is recommended to set this value to the maximum number of alerts
        "max_allowed_cache_size": 20000,
        # default limit for listing alert configs
        "default_list_alert_configs_limit": 2000,
    },
    "auth_with_client_id": {
        "enabled": False,
        "request_timeout": 5,
    },
    "services": {
        # The running service name. One of: "api", "alerts"
        "service_name": "api",
        "hydra": {
            # Comma separated list of services to run on the instance.
            # Currently, this is only considered when the service_name is "api".
            # "*" starts all services on the same instance,
            # other options are considered as running only the api service.
            "services": "*",
        },
    },
    "notifications": {
        "smtp": {
            "config_secret_name": "mlrun-smtp-config",
            "refresh_interval": "30",
        }
    },
    "system_id": "",
}
_is_running_as_api = None


def is_running_as_api():
    # MLRUN_IS_API_SERVER is set when running the api server which is being done through the CLI command mlrun db
    global _is_running_as_api

    if _is_running_as_api is None:
        _is_running_as_api = os.getenv("MLRUN_IS_API_SERVER", "false").lower() == "true"

    return _is_running_as_api


class Config:
    _missing = object()

    def __init__(self, cfg=None):
        cfg = {} if cfg is None else cfg

        # Can't use self._cfg = cfg → infinite recursion
        object.__setattr__(self, "_cfg", cfg)

    def __getattr__(self, attr):
        val = self._cfg.get(attr, self._missing)
        if val is self._missing:
            raise AttributeError(attr)

        if isinstance(val, Mapping):
            return self.__class__(val)
        return val

    def __deepcopy__(self, memo):
        cls = self.__class__
        # create a new Config without calling __init__ (avoid recursion)
        result = cls.__new__(cls)
        # manually deep-copy _cfg
        object.__setattr__(result, "_cfg", copy.deepcopy(self._cfg, memo))
        return result

    def __setattr__(self, attr, value):
        self._cfg[attr] = value

    def __dir__(self):
        return list(self._cfg) + dir(self.__class__)

    def __repr__(self):
        name = self.__class__.__name__
        return f"{name}({self._cfg!r})"

    def __iter__(self):
        if isinstance(self._cfg, Mapping):
            return self._cfg.__iter__()

    def items(self):
        if isinstance(self._cfg, Mapping):
            return iter(self._cfg.items())

    def keys(self):
        if isinstance(self._cfg, Mapping):
            return iter(self.data.keys())

    def values(self):
        if isinstance(self._cfg, Mapping):
            return iter(self.data.values())

    def update(self, cfg, skip_errors=False):
        for key, value in cfg.items():
            if hasattr(self, key):
                if isinstance(value, dict):
                    # ignore the `skip_errors` flag here
                    # if the key does not align with what mlrun config expects it is a user
                    # input error that can lead to unexpected behavior.
                    # raise the exception to ensure configuration is loaded correctly and do not
                    # ignore any errors.
                    config_value = getattr(self, key)
                    try:
                        config_value.update(value)
                    except AttributeError as exc:
                        if not isinstance(config_value, (dict, Config)):
                            raise ValueError(
                                f"Can not update `{key}` config. "
                                f"Expected a configuration but received {type(value)}"
                            ) from exc
                        raise exc
                else:
                    try:
                        setattr(self, key, value)
                    except mlrun.errors.MLRunRuntimeError as exc:
                        if not skip_errors:
                            raise exc
                        print(
                            f"Warning, failed to set config key {key}={value}, {mlrun.errors.err_to_str(exc)}"
                        )

    def dump_yaml(self, stream=None):
        return yaml.dump(self._cfg, stream, default_flow_style=False)

    @classmethod
    def from_dict(cls, dict_):
        return cls(copy.deepcopy(dict_))

    @staticmethod
    def get_build_args():
        build_args = {}
        if config.httpdb.builder.build_args:
            build_args_json = base64.b64decode(
                config.httpdb.builder.build_args
            ).decode()
            build_args = json.loads(build_args_json)

        return build_args

    @staticmethod
    def is_pip_ca_configured():
        return (
            config.httpdb.builder.pip_ca_secret_name
            and config.httpdb.builder.pip_ca_secret_key
            and config.httpdb.builder.pip_ca_path
        )

    @staticmethod
    def get_default_hub_source_url_prefix(object_type) -> str:
        default_source = config.hub.default_source
        return f"{default_source.url}/{object_type}/{default_source.channel}/"

    @staticmethod
    def decode_base64_config_and_load_to_object(
        attribute_path: str, expected_type=dict
    ):
        """
        decodes and loads the config attribute to expected type

        :param attribute_path: the path in the default_config e.g. preemptible_nodes.node_selector
        :param expected_type: the object type valid values are : `dict`, `list` etc...
        :return: the expected type instance
        """
        attributes = attribute_path.split(".")
        raw_attribute_value = config
        for part in attributes:
            try:
                raw_attribute_value = raw_attribute_value.__getattr__(part)
            except AttributeError:
                raise mlrun.errors.MLRunNotFoundError(
                    "Attribute does not exist in config"
                )
        # There is a bug in the installer component in iguazio system that causes the configured value to be base64 of
        # null (without conditioning it we will end up returning None instead of empty dict)
        if raw_attribute_value and raw_attribute_value != "bnVsbA==":
            try:
                decoded_attribute_value = base64.b64decode(raw_attribute_value).decode()
            except Exception:
                raise mlrun.errors.MLRunInvalidArgumentTypeError(
                    f"Unable to decode {attribute_path}"
                )
            parsed_attribute_value = json.loads(decoded_attribute_value)
            if not isinstance(parsed_attribute_value, expected_type):
                raise mlrun.errors.MLRunInvalidArgumentTypeError(
                    f"Expected type {expected_type}, got {type(parsed_attribute_value)}"
                )
            return parsed_attribute_value
        return expected_type()

    def get_default_function_node_selector(self) -> dict:
        return self.decode_base64_config_and_load_to_object(
            "default_function_node_selector", dict
        )

    def get_preemptible_node_selector(self) -> dict:
        return self.decode_base64_config_and_load_to_object(
            "preemptible_nodes.node_selector", dict
        )

    def get_preemptible_tolerations(self) -> list:
        return self.decode_base64_config_and_load_to_object(
            "preemptible_nodes.tolerations", list
        )

    def get_default_function_security_context(self) -> dict:
        return self.decode_base64_config_and_load_to_object(
            "function.spec.security_context.default", dict
        )

    def is_preemption_nodes_configured(self):
        return (
            self.get_preemptible_tolerations() or self.get_preemptible_node_selector()
        )

    @staticmethod
    def get_valid_function_priority_class_names():
        valid_function_priority_class_names = []
        if not config.valid_function_priority_class_names:
            return valid_function_priority_class_names

        # Manually ensure we have only unique values because we want to keep the order and using a set would lose it
        for priority_class_name in config.valid_function_priority_class_names.split(
            ","
        ):
            if priority_class_name not in valid_function_priority_class_names:
                valid_function_priority_class_names.append(priority_class_name)
        return valid_function_priority_class_names

    @staticmethod
    def is_running_on_iguazio() -> bool:
        return config.igz_version is not None and config.igz_version != ""

    @staticmethod
    def get_security_context_enrichment_group_id(user_unix_id: int) -> int:
        enrichment_group_id = int(
            config.function.spec.security_context.enrichment_group_id
        )

        # if enrichment group id is -1 we set group id to user unix id
        if enrichment_group_id == -1:
            if user_unix_id is None:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "User unix id is required to populate group id when enrichment group id is -1."
                    "See mlrun.config.function.spec.security_context.enrichment_group_id for more details."
                )
            return user_unix_id

        return enrichment_group_id

    @staticmethod
    def get_parsed_igz_version() -> typing.Optional[semver.VersionInfo]:
        if not config.igz_version:
            return None
        try:
            parsed_version = semver.VersionInfo.parse(config.igz_version)
            return parsed_version
        except ValueError:
            # iguazio version is semver compatible only from 3.2, before that it will be something
            # like 3.0_b177_20210806003728
            semver_compatible_igz_version = config.igz_version.split("_")[0]
            return semver.VersionInfo.parse(f"{semver_compatible_igz_version}.0")

    def verify_security_context_enrichment_mode_is_allowed(self):
        if (
            config.function.spec.security_context.enrichment_mode
            == mlrun.common.schemas.function.SecurityContextEnrichmentModes.disabled
        ):
            return

        igz_version = self.get_parsed_igz_version()
        if not igz_version:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Unable to determine if security context enrichment mode is allowed. Missing iguazio version"
            )

        if igz_version < semver.VersionInfo.parse("3.5.1-b1"):
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Security context enrichment mode enabled (override/retain) "
                f"is not allowed for iguazio version: {igz_version} < 3.5.1"
            )

    def validate_object_retentions(self):
        for table_name, retention_days in self.object_retentions.items():
            if retention_days < 7 and not os.getenv("PARTITION_INTERVAL"):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"{table_name} partition interval must be greater than a week"
                )
            elif retention_days > 53 * 7:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"{table_name} partition interval must be less than a year"
                )

    def resolve_chief_api_url(self) -> str:
        if self.httpdb.clusterization.chief.url:
            return self.httpdb.clusterization.chief.url
        if not self.httpdb.clusterization.chief.service:
            raise mlrun.errors.MLRunNotFoundError(
                "For resolving chief url, chief service name must be provided"
            )
        if self.namespace is None:
            raise mlrun.errors.MLRunNotFoundError(
                "For resolving chief url, namespace must be provided"
            )

        chief_api_url = f"http://{self.httpdb.clusterization.chief.service}.{self.namespace}.svc.cluster.local"
        if config.httpdb.clusterization.chief.port:
            chief_api_url = f"{chief_api_url}:{self.httpdb.clusterization.chief.port}"

        self.httpdb.clusterization.chief.url = chief_api_url
        return self.httpdb.clusterization.chief.url

    @staticmethod
    def internal_labels():
        return mlrun.common.constants.MLRunInternalLabels.all()

    @staticmethod
    def get_storage_auto_mount_params():
        auto_mount_params = {}
        if config.storage.auto_mount_params:
            try:
                auto_mount_params = base64.b64decode(
                    config.storage.auto_mount_params, validate=True
                ).decode()
                auto_mount_params = json.loads(auto_mount_params)
            except binascii.Error:
                # Importing here to avoid circular dependencies
                from .utils import list2dict

                # String wasn't base64 encoded. Parse it using a 'p1=v1,p2=v2' format.
                mount_params = config.storage.auto_mount_params.split(",")
                auto_mount_params = list2dict(mount_params)
        if not isinstance(auto_mount_params, dict):
            raise TypeError(
                f"data in storage.auto_mount_params does not resolve to a dictionary: {auto_mount_params}"
            )

        return auto_mount_params

    def get_default_function_pod_resources(
        self, with_gpu_requests=False, with_gpu_limits=False
    ):
        resources = {}
        resource_requirements = ["requests", "limits"]
        for requirement in resource_requirements:
            with_gpu = (
                with_gpu_requests if requirement == "requests" else with_gpu_limits
            )
            resources[requirement] = (
                self.get_default_function_pod_requirement_resources(
                    requirement, with_gpu
                )
            )
        return resources

    def resolve_runs_monitoring_missing_runtime_resources_debouncing_interval(self):
        return (
            float(self.monitoring.runs.missing_runtime_resources_debouncing_interval)
            if self.monitoring.runs.missing_runtime_resources_debouncing_interval
            else float(config.monitoring.runs.interval) * 2.0
        )

    @staticmethod
    def get_default_function_pod_requirement_resources(
        requirement: str, with_gpu: bool = True
    ):
        """
        :param requirement: kubernetes requirement resource one of the following : requests, limits
        :param with_gpu: whether to return requirement resources with nvidia.com/gpu field (e.g. you cannot specify
         GPU requests without specifying GPU limits) https://kubernetes.io/docs/tasks/manage-gpus/scheduling-gpus/
        :return: a dict containing the defaults resources (cpu, memory, nvidia.com/gpu)
        """
        resources: dict = copy.deepcopy(config.default_function_pod_resources.to_dict())
        gpu_type = "nvidia.com/gpu"
        gpu = "gpu"
        resource_requirement = resources.get(requirement, {})
        resource_requirement.setdefault(gpu)
        if with_gpu:
            resource_requirement[gpu_type] = resource_requirement.pop(gpu)
        else:
            resource_requirement.pop(gpu)
        return resource_requirement

    def force_api_gateway_ssl_redirect(self):
        """
        Get the default value for the ssl_redirect configuration.
        In Iguazio we always want to redirect to HTTPS, in other cases we don't.

        :return: True if we should redirect to HTTPS, False otherwise.
        """
        return self.is_running_on_iguazio()

    @staticmethod
    def get_run_retry_staleness_threshold_timedelta() -> timedelta:
        """
        Get the staleness threshold in timedelta for run retries.
        This is used to determine if a run is stale and should be retried.

        :return: The staleness threshold in timedelta.
        """
        staleness_threshold = int(
            mlrun.mlconf.monitoring.runs.retry.staleness_threshold
        )
        return timedelta(minutes=staleness_threshold)

    def to_dict(self):
        return copy.deepcopy(self._cfg)

    @staticmethod
    def reload():
        _populate()

    @property
    def version(self):
        # importing here to avoid circular dependency
        from mlrun.utils.version import Version

        return Version().get()["version"]

    @staticmethod
    def resolve_ui_url():
        # ui_url is deprecated in favor of the ui.url (we created the ui block)
        # since the config class is used in a "recursive" way, we can't use property like we used in other places
        # since the property will need to be url, which exists in other structs as well
        return config.ui.url or config.ui_url

    def is_api_running_on_k8s(self):
        # determine if the API service is attached to K8s cluster
        # when there is a cluster the .namespace is set
        return bool(mlrun.mlconf.namespace)

    def is_nuclio_detected(self):
        # determine is Nuclio service is detected, when the nuclio_version is not set
        return bool(mlrun.mlconf.nuclio_version)

    def use_nuclio_mock(self, force_mock=None):
        # determine if to use Nuclio mock service
        mock_nuclio = mlrun.mlconf.mock_nuclio_deployment
        if mock_nuclio and mock_nuclio == "auto":
            mock_nuclio = not mlrun.mlconf.is_nuclio_detected()
        return True if mock_nuclio and force_mock is None else force_mock

    def get_v3io_access_key(self) -> typing.Optional[str]:
        # Get v3io access key from the environment
        return os.getenv("V3IO_ACCESS_KEY")

    def get_model_monitoring_file_target_path(
        self,
        project: str,
        kind: str,
        target: typing.Literal["online", "offline"] = "online",
        artifact_path: typing.Optional[str] = None,
        function_name: typing.Optional[str] = None,
        **kwargs,
    ) -> str:
        """Get the full path from the configuration based on the provided project and kind.

        :param project:         Project name.
        :param kind:            Kind of target path (e.g. events, log_stream, endpoints, etc.)
        :param target:          Can be either online or offline. If the target is online, then we try to get a specific
                                path for the provided kind. If it doesn't exist, use the default path.
                                If the target path is offline and the offline path is already a full path in the
                                configuration, then the result will be that path as-is. If the offline path is a
                                relative path, then the result will be based on the project artifact path and the
                                offline relative path. If project artifact path wasn't provided, then we use MLRun
                                artifact path instead.
        :param artifact_path:   Optional artifact path that will be used as a relative path. If not provided, the
                                relative artifact path will be taken from the global MLRun artifact path.
        :param function_name:    Application name, None for model_monitoring_stream.

        :return:                Full configured path for the provided kind.
        """

        if target != "offline":
            store_prefix_dict = (
                mlrun.mlconf.model_endpoint_monitoring.store_prefixes.to_dict()
            )
            if store_prefix_dict.get(kind):
                # Target exist in store prefix and has a valid string value
                return store_prefix_dict[kind].format(project=project, **kwargs)
            if (
                function_name
                and function_name
                != mlrun.common.schemas.model_monitoring.constants.MonitoringFunctionNames.STREAM
                and function_name
                != mlrun.common.schemas.model_monitoring.constants.MonitoringFunctionNames.APPLICATION_CONTROLLER
            ):
                return mlrun.mlconf.model_endpoint_monitoring.store_prefixes.user_space.format(
                    project=project,
                    kind=kind
                    if function_name is None
                    else f"{kind}-{function_name.lower()}-v1",
                )
            elif (
                kind == "stream"
                and function_name
                != mlrun.common.schemas.model_monitoring.constants.MonitoringFunctionNames.APPLICATION_CONTROLLER
            ):
                return mlrun.mlconf.model_endpoint_monitoring.store_prefixes.user_space.format(
                    project=project,
                    kind=f"{kind}-v1",
                )
            elif (
                function_name
                == mlrun.common.schemas.model_monitoring.constants.MonitoringFunctionNames.APPLICATION_CONTROLLER
                and kind == "stream"
            ):
                return mlrun.mlconf.model_endpoint_monitoring.store_prefixes.default.format(
                    project=project,
                    kind=f"{kind}-{function_name.lower()}-v1",
                )

            return mlrun.mlconf.model_endpoint_monitoring.store_prefixes.default.format(
                project=project,
                kind=kind,
            )

        # Get the current offline path from the configuration
        file_path = mlrun.mlconf.model_endpoint_monitoring.offline_storage_path.format(
            project=project, kind=kind
        )

        # Absolute path
        if any(value in file_path for value in ["://", ":///"]) or os.path.isabs(
            file_path
        ):
            return file_path

        # Relative path
        else:
            artifact_path = artifact_path or config.artifact_path
            if artifact_path[-1] != "/":
                artifact_path += "/"

            return mlrun.utils.helpers.template_artifact_path(
                artifact_path=artifact_path + file_path, project=project
            )

    def is_ce_mode(self) -> bool:
        # True if the setup is in CE environment
        return isinstance(mlrun.mlconf.ce, mlrun.config.Config) and any(
            ver in mlrun.mlconf.ce.mode for ver in ["lite", "full"]
        )

    def is_explicit_ack_enabled(self) -> bool:
        return self.httpdb.nuclio.explicit_ack == "enabled" and (
            not self.nuclio_version
            or semver.VersionInfo.parse(self.nuclio_version)
            >= semver.VersionInfo.parse("1.12.10")
        )


# Global configuration
config = Config.from_dict(default_config)


def _populate(skip_errors=False):
    """Populate configuration from config file (if exists in environment) and
    from environment variables.

    populate will run only once, after first call it does nothing.
    """
    global _loaded

    with _load_lock:
        _do_populate(skip_errors=skip_errors)


def _do_populate(env=None, skip_errors=False):
    global config

    if not os.environ.get("MLRUN_IGNORE_ENV_FILE"):
        if "MLRUN_ENV_FILE" in os.environ:
            env_file = os.path.expanduser(os.environ["MLRUN_ENV_FILE"])
            dotenv.load_dotenv(env_file, override=True)
        else:
            env_file = os.path.expanduser(default_env_file)
            if os.path.isfile(env_file):
                dotenv.load_dotenv(env_file, override=True)

    if not config:
        config = Config.from_dict(default_config)
    else:
        config.update(default_config)
    config_path = os.environ.get(env_file_key)
    if config_path:
        with open(config_path) as fp:
            data = yaml.safe_load(fp)

        if not isinstance(data, dict):
            raise TypeError(f"configuration in {config_path} not a dict")

        config.update(data, skip_errors=skip_errors)

    data = read_env(env)
    if data:
        config.update(data, skip_errors=skip_errors)

    _configure_ssl_verification(config.httpdb.http.verify)
    _validate_config(config)

    if config.dbpath:
        from mlrun.db import get_run_db

        # when dbpath is set we want to connect to it which will sync configuration from it to the client
        get_run_db(config.dbpath, force_reconnect=True)


def _validate_config(config):
    try:
        limits_gpu = config.default_function_pod_resources.limits.gpu
        requests_gpu = config.default_function_pod_resources.requests.gpu
        _verify_gpu_requests_and_limits(
            requests_gpu=requests_gpu,
            limits_gpu=limits_gpu,
        )
    except AttributeError:
        pass

    config.verify_security_context_enrichment_mode_is_allowed()
    config.validate_object_retentions()


def _verify_gpu_requests_and_limits(
    requests_gpu: typing.Optional[str] = None, limits_gpu: typing.Optional[str] = None
):
    # https://kubernetes.io/docs/tasks/manage-gpus/scheduling-gpus/
    if requests_gpu and not limits_gpu:
        raise mlrun.errors.MLRunConflictError(
            "You cannot specify GPU requests without specifying limits"
        )
    if requests_gpu and limits_gpu and requests_gpu != limits_gpu:
        raise mlrun.errors.MLRunConflictError(
            f"When specifying both GPU requests and limits these two values must be equal, "
            f"requests_gpu={requests_gpu}, limits_gpu={limits_gpu}"
        )


def _convert_resources_to_str(config: typing.Optional[dict] = None):
    resources_types = ["cpu", "memory", "gpu"]
    resource_requirements = ["requests", "limits"]
    if not config.get("default_function_pod_resources"):
        return
    for requirement in resource_requirements:
        resource_requirement = config.get("default_function_pod_resources").get(
            requirement
        )
        if not resource_requirement:
            continue
        for resource_type in resources_types:
            value = resource_requirement.setdefault(resource_type, None)
            if value is None:
                continue
            resource_requirement[resource_type] = str(value)


def _configure_ssl_verification(verify_ssl: bool) -> None:
    """Configure SSL verification warnings based on the setting."""
    if not verify_ssl:
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    else:
        # If the user changes the `verify` setting to `True` at runtime using `mlrun.set_env_from_file` after
        # importing `mlrun`, we need to reload the `mlrun` configuration and enable this warning.
        warnings.simplefilter("default", urllib3.exceptions.InsecureRequestWarning)


def read_env(env=None, prefix=env_prefix):
    """Read configuration from environment"""
    env = os.environ if env is None else env

    config = {}
    for key, value in env.items():
        if not key.startswith(env_prefix) or key == env_file_key:
            continue
        try:
            value = json.loads(value)  # values can be JSON encoded
        except ValueError:
            pass  # Leave as string
        key = key[len(env_prefix) :]  # Trim MLRUN_
        path = key.lower().split("__")  # 'A__B' → ['a', 'b']
        cfg = config
        while len(path) > 1:
            name, *path = path
            cfg = cfg.setdefault(name, {})
        cfg[path[0]] = value

    env_dbpath = env.get("MLRUN_DBPATH", "")
    # expected format: https://mlrun-api.tenant.default-tenant.app.some-system.some-namespace.com
    is_remote_mlrun = (
        env_dbpath.startswith("https://mlrun-api.") and "tenant." in env_dbpath
    )

    # It's already a standard to set this env var to configure the v3io api, so we're supporting it (instead
    # of MLRUN_V3IO_API), in remote usage this can be auto detected from the DBPATH
    v3io_api = env.get("V3IO_API")
    if v3io_api:
        config["v3io_api"] = v3io_api
    elif is_remote_mlrun:
        # in remote mlrun we can't use http, so we'll use https
        config["v3io_api"] = env_dbpath.replace("https://mlrun-api.", "https://webapi.")

    # It's already a standard to set this env var to configure the v3io framesd, so we're supporting it (instead
    # of MLRUN_V3IO_FRAMESD), in remote usage this can be auto detected from the DBPATH
    v3io_framesd = env.get("V3IO_FRAMESD")
    if v3io_framesd:
        config["v3io_framesd"] = v3io_framesd
    elif is_remote_mlrun:
        config["v3io_framesd"] = env_dbpath.replace(
            "https://mlrun-api.", "https://framesd."
        )

    uisvc = env.get("MLRUN_UI_SERVICE_HOST")
    igz_domain = env.get("IGZ_NAMESPACE_DOMAIN")

    # workaround to try and detect IGZ domain
    if not igz_domain and "MLRUN_HTTPDB__BUILDER__DOCKER_REGISTRY" in env:
        registry = env["MLRUN_HTTPDB__BUILDER__DOCKER_REGISTRY"]
        if registry.startswith("docker-registry.default-tenant"):
            igz_domain = registry[len("docker-registry.") :]
            if ":" in igz_domain:
                igz_domain = igz_domain[: igz_domain.rfind(":")]
            env["IGZ_NAMESPACE_DOMAIN"] = igz_domain

    # workaround wrongly sqldb dsn in 2.8
    if (
        config.get("httpdb", {}).get("dsn")
        == "sqlite:///mlrun.sqlite3?check_same_thread=false"
    ):
        config["httpdb"]["dsn"] = "sqlite:////mlrun/db/mlrun.db?check_same_thread=false"

    # "disabled" is the helm chart default value, we don't want that value to be set cause when this value is set we
    # use it in calls to the Nuclio package, and when the Nuclio package receives a value it simply uses it, and
    # obviously "disabled" is not the right address.. when the Nuclio package doesn't receive a value it doing "best
    # effort" to try and determine the URL, we want this "best effort" so overriding the "disabled" value
    if config.get("nuclio_dashboard_url") == "disabled":
        config["nuclio_dashboard_url"] = ""

    if uisvc and not config.get("ui_url"):
        if igz_domain:
            config["ui_url"] = f"https://mlrun-ui.{igz_domain}"

    if log_level := config.get("log_level"):
        import mlrun.utils.logger

        # logger created (because of imports mess) before the config is loaded (in tests), therefore we're changing its
        # level manually
        mlrun.utils.logger.set_logger_level(log_level)

    if log_formatter_name := config.get("log_formatter"):
        import mlrun.utils.logger

        log_formatter = mlrun.utils.resolve_formatter_by_kind(
            mlrun.utils.FormatterKinds(log_formatter_name)
        )
        current_handler = mlrun.utils.logger.get_handler("default")
        current_formatter_name = current_handler.formatter.__class__.__name__
        desired_formatter_name = log_formatter.__name__
        if current_formatter_name != desired_formatter_name:
            current_handler.setFormatter(log_formatter())

    # The default function pod resource values are of type str; however, when reading from environment variable numbers,
    # it converts them to type int if contains only number, so we want to convert them to str.
    _convert_resources_to_str(config)

    # If the environment variable MLRUN_HTTPDB__HTTP__VERIFY is set, we ensure SSL verification settings take precedence
    # by moving the 'httpdb' configuration to the beginning of the config dictionary.
    # This ensures that SSL verification is applied before other settings.
    if "MLRUN_HTTPDB__HTTP__VERIFY" in env:
        httpdb = config.pop("httpdb", None)
        if httpdb:
            config = {"httpdb": httpdb, **config}

    return config


# populate config, skip errors when setting the config attributes and issue warnings instead
# this is to avoid failure when doing `import mlrun` and the dbpath (API service) is incorrect or down
_populate(skip_errors=True)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import warnings
from pprint import pprint
from time import sleep
from typing import Optional

from .iguazio import (
    V3ioStreamClient,
    add_or_refresh_credentials,
    is_iguazio_session_cookie,
)


# TODO: Remove in 1.11.0
class _DeprecationHelper:
    """A helper class to deprecate old schemas"""

    def __init__(self, new_target: str, version="1.8.0"):
        self._new_target = new_target
        self._version = version

    def __call__(self, *args, **kwargs):
        self._warn()
        return self._lazy_load()(*args, **kwargs)

    def __getattr__(self, attr):
        self._warn()
        return getattr(self._lazy_load(), attr)

    def _lazy_load(self, *args, **kwargs):
        import mlrun.runtimes.mounts as mlrun_mounts

        return getattr(mlrun_mounts, self._new_target)

    def _warn(self):
        warnings.warn(
            f"mlrun.platforms.{self._new_target} is deprecated since version {self._version}, "
            f"and will be removed in 1.11.0. Use mlrun.runtimes.mounts.{self._new_target} instead.",
            FutureWarning,
        )


# TODO: Remove in 1.11.0
# For backwards compatibility
VolumeMount = _DeprecationHelper("VolumeMount")
auto_mount = _DeprecationHelper("auto_mount")
mount_configmap = _DeprecationHelper("mount_configmap")
mount_hostpath = _DeprecationHelper("mount_hostpath")
mount_pvc = _DeprecationHelper("mount_pvc")
mount_s3 = _DeprecationHelper("mount_s3")
mount_secret = _DeprecationHelper("mount_secret")
mount_v3io = _DeprecationHelper("mount_v3io")
set_env_variables = _DeprecationHelper("set_env_variables")
v3io_cred = _DeprecationHelper("v3io_cred")
# eof 'For backwards compatibility'


def watch_stream(
    url,
    shard_ids: Optional[list] = None,
    seek_to: Optional[str] = None,
    interval=None,
    is_json=False,
    **kwargs,
):
    """watch on a v3io stream and print data every interval

    example::

        watch_stream("v3io:///users/admin/mystream")

    :param url:        stream url
    :param shard_ids:  range or list of shard IDs
    :param seek_to:    where to start/seek ('EARLIEST', 'LATEST', 'TIME', 'SEQUENCE')
    :param interval:   watch interval time in seconds, 0 to run once and return
    :param is_json:    indicate the payload is json (will be deserialized)
    """
    interval = 3 if interval is None else interval
    shard_ids = shard_ids or [0]
    if isinstance(shard_ids, int):
        shard_ids = [shard_ids]
    watchers = [
        V3ioStreamClient(url, shard_id, seek_to, **kwargs)
        for shard_id in list(shard_ids)
    ]
    while True:
        for watcher in watchers:
            records = watcher.get_records()
            for record in records:
                print(
                    f"{watcher.url}:{watcher.shard_id} (#{record.sequence_number}) >> "
                )
                data = json.loads(record.data) if is_json else record.data.decode()
                pprint(data)
        if interval <= 0:
            break
        sleep(interval)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import mlrun
from mlrun.artifacts import Artifact

from .plan import Plan
from .utils import LoggingMode


class Producer:
    """
    Class for handling production of artifact plans during a run.
    """

    def __init__(
        self,
        context: mlrun.MLClientCtx = None,
        plans: Optional[list[Plan]] = None,
    ):
        """
        Initialize a producer with the given plans. The producer will log the produced artifacts using the given
        context.

        :param context: The context to log with.
        :param plans:   The plans the producer will manage.
        """
        # Store the context and plans:
        self._context = context
        self._plans = plans if plans is not None else []

        # Set up the logger's mode (default:  Training):
        self._mode = LoggingMode.TRAINING

        # Prepare the dictionaries to hold the artifacts. Once they are logged they will be moved from one to another:
        self._logged_artifacts = {}  # type: Dict[str, Artifact]
        self._not_logged_artifacts = {}  # type: Dict[str, Artifact]

    @property
    def mode(self) -> LoggingMode:
        """
        Get the logger's mode.

        :return: The logger mode.
        """
        return self._mode

    @property
    def context(self) -> mlrun.MLClientCtx:
        """
        Get the logger's MLRun context.

        :return: The logger's MLRun context.
        """
        return self._context

    @property
    def artifacts(self) -> dict[str, Artifact]:
        """
        Get the logged artifacts.

        :return: The logged artifacts.
        """
        return self._logged_artifacts

    def set_mode(self, mode: LoggingMode):
        """
        Set the producer's mode.

        :param mode: The mode to set.
        """
        self._mode = mode

    def set_context(self, context: mlrun.MLClientCtx):
        """
        Set the context this logger will log with.

        :param context: The to be set MLRun context.
        """
        self._context = context

    def set_plans(self, plans: list[Plan]):
        """
        Update the plans of this logger to the given list of plans here.

        :param plans: The list of plans to override the current one.
        """
        self._plans = plans

    def produce_stage(self, stage, **kwargs):
        """
        Produce the artifacts ready at the given stage and log them.

        :param stage:  The current stage to log at.
        :param kwargs: All the required produce arguments to pass onto the plans.
        """
        # Produce all the artifacts according to the given stage:
        self._produce_artifacts(stage=stage, **kwargs)

        # Log if a context is available:
        if self._context is not None:
            # Log the artifacts in queue:
            self._log_artifacts()
            # Commit:
            self._context.commit(completed=False)

    def _produce_artifacts(self, stage, **kwargs):
        """
        Go through the plans and check if they are ready to be produced in the given stage of the run. If they are,
        the logger will pass all the arguments to the 'plan.produce' method and collect the returned artifact.

        :param stage:            The stage to produce the artifact to check if its ready.
        :param kwargs:           All of the required produce arguments to pass onto the plans.
        """
        # Initialize a new list of plans for all the plans that will still need to be produced:
        plans = []

        # Go ver the plans to produce their artifacts:
        for plan in self._plans:
            # Check if the plan is ready:
            if plan.is_ready(stage=stage):
                # Produce the artifact:
                self._not_logged_artifacts = {
                    **self._not_logged_artifacts,
                    **plan.produce(**kwargs),
                }
                # If the plan should not be produced again, continue to the next one so it won't be collected:
                if not plan.is_reproducible():
                    continue
            # Collect the plan to produce it later (or again if reproducible):
            plans.append(plan)

        # Clear the old plans:
        self._plans = plans

        # Add evaluation prefix if in Evaluation mode:
        if self._mode == LoggingMode.EVALUATION:
            self._not_logged_artifacts = {
                f"evaluation-{key}": value
                for key, value in self._not_logged_artifacts.items()
            }
            for artifact in self._not_logged_artifacts.values():
                artifact.key = f"evaluation-{artifact.key}"

    def _log_artifacts(self):
        """
        Log the produced plans artifacts using the logger's context.
        """
        # Use the context to log each artifact:
        for artifact in self._not_logged_artifacts.values():
            self._context.log_artifact(artifact)

        # Collect the logged artifacts:
        self._logged_artifacts = {
            **self._logged_artifacts,
            **self._not_logged_artifacts,
        }

        # Clean the not logged artifacts dictionary:
        self._not_logged_artifacts = {}



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from datetime import datetime
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse

import oss2
from fsspec.registry import get_filesystem_class

import mlrun.errors

from .base import DataStore, FileStats, make_datastore_schema_sanitizer


class OSSStore(DataStore):
    using_bucket = True

    def __init__(
        self, parent, schema, name, endpoint="", secrets: Optional[dict] = None
    ):
        super().__init__(parent, name, schema, endpoint, secrets)
        # will be used in case user asks to assume a role and work through fsspec

        access_key_id = self._get_secret_or_env("ALIBABA_ACCESS_KEY_ID")
        secret_key = self._get_secret_or_env("ALIBABA_SECRET_ACCESS_KEY")
        endpoint_url = self._get_secret_or_env("ALIBABA_ENDPOINT_URL")
        if access_key_id and secret_key and endpoint_url:
            self.auth = oss2.Auth(access_key_id, secret_key)
            self.endpoint_url = endpoint_url
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "missing ALIBABA_ACCESS_KEY_ID or ALIBABA_SECRET_ACCESS_KEY ALIBABA_ENDPOINT_URL in environment"
            )

    @property
    def filesystem(self):
        """return fsspec file system object, if supported"""
        if self._filesystem:
            return self._filesystem
        try:
            import ossfs  # noqa
        except ImportError as exc:
            raise ImportError("ALIBABA ossfs not installed") from exc
        filesystem_class = get_filesystem_class(protocol=self.kind)
        self._filesystem = make_datastore_schema_sanitizer(
            filesystem_class,
            using_bucket=self.using_bucket,
            **self.get_storage_options(),
        )
        return self._filesystem

    def get_storage_options(self):
        res = dict(
            endpoint=self._get_secret_or_env("ALIBABA_ENDPOINT_URL"),
            key=self._get_secret_or_env("ALIBABA_ACCESS_KEY_ID"),
            secret=self._get_secret_or_env("ALIBABA_SECRET_ACCESS_KEY"),
        )
        return self._sanitize_options(res)

    def get_bucket_and_key(self, key):
        path = self._join(key)[1:]
        return self.endpoint, path

    def upload(self, key, src_path):
        bucket, key = self.get_bucket_and_key(key)
        oss = oss2.Bucket(self.auth, self.endpoint_url, bucket)
        oss.put_object(key, open(src_path, "rb"))

    def get(self, key, size=None, offset=0):
        bucket, key = self.get_bucket_and_key(key)
        oss = oss2.Bucket(self.auth, self.endpoint_url, bucket)
        if size or offset:
            return oss.get_object(key, byte_range=self.get_range(size, offset)).read()
        return oss.get_object(key).read()

    def put(self, key, data, append=False):
        data, _ = self._prepare_put_data(data, append)
        bucket, key = self.get_bucket_and_key(key)
        oss = oss2.Bucket(self.auth, self.endpoint_url, bucket)
        oss.put_object(key, data)

    def stat(self, key):
        bucket, key = self.get_bucket_and_key(key)
        oss = oss2.Bucket(self.auth, self.endpoint_url, bucket)
        obj = oss.get_object_meta(key)
        size = obj.content_length
        modified = datetime.fromtimestamp(obj.last_modified)
        return FileStats(size, time.mktime(modified.timetuple()))

    def listdir(self, key):
        remote_path = self._convert_key_to_remote_path(key)
        if self.filesystem.isfile(remote_path):
            return key
        remote_path = f"{remote_path}/**"
        files = self.filesystem.glob(remote_path)
        key_length = len(key)
        files = [
            f.split("/", 1)[1][key_length:] for f in files if len(f.split("/")) > 1
        ]
        return files

    def delete(self, key):
        bucket, key = self.get_bucket_and_key(key)
        oss = oss2.Bucket(self.auth, self.endpoint_url, bucket)
        oss.delete_object(key)

    def _convert_key_to_remote_path(self, key):
        key = key.strip("/")
        schema = urlparse(key).scheme
        #  if called without passing dataitem - like in fset.purge_targets,
        #  key will include schema.
        if not schema:
            key = Path(self.endpoint, key).as_posix()
        return key

    @staticmethod
    def get_range(size, offset):
        if size:
            return [offset, size]
        return [offset, None]



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC

import lightgbm as lgb

from ..._common import MLRunInterface
from ..utils import LGBMTypes


class LGBMBoosterMLRunInterface(MLRunInterface, ABC):
    """
    Interface for adding MLRun features for LightGBM models (Booster API).
    """

    _PROPERTIES = {
        "model_handler": None,  # type: MLModelHandler
    }

    @classmethod
    def add_interface(
        cls,
        obj: lgb.Booster,
        restoration: LGBMTypes.MLRunInterfaceRestorationType = None,
    ):
        """
        Enrich the object with this interface properties, methods and functions, so it will have this LightGBM MLRun's
        features.

        :param obj:         The object to enrich his interface.
        :param restoration: Restoration information tuple as returned from 'remove_interface' in order to add the
                            interface in a certain state.
        """
        super().add_interface(obj=obj, restoration=restoration)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import datetime
import os
import traceback
import typing
from concurrent.futures import ThreadPoolExecutor

import mlrun.common.constants as mlrun_constants
import mlrun.common.runtimes.constants as runtimes_constants
import mlrun.common.schemas
import mlrun.config
import mlrun.db.base
import mlrun.errors
import mlrun.lists
import mlrun.model
import mlrun.utils.helpers
import mlrun.utils.notifications.notification as notification_module
import mlrun.utils.notifications.notification.base as base
from mlrun.utils import Workflow, logger
from mlrun.utils.condition_evaluator import evaluate_condition_in_separate_process


class _NotificationPusherBase:
    def _push(
        self, sync_push_callback: typing.Callable, async_push_callback: typing.Callable
    ):
        if mlrun.utils.helpers.is_running_in_jupyter_notebook():
            # Running in Jupyter notebook.
            # In this case, we need to create a new thread, run a separate event loop in
            # that thread, and use it instead of the main_event_loop.
            # This is necessary because Jupyter Notebook has its own event loop,
            # but it runs in the main thread. As long as a cell is running,
            # the event loop will not execute properly
            self._run_coroutine_in_jupyter_notebook(
                coroutine_method=async_push_callback
            )
        else:
            # Either running in mlrun api or sdk. in case of mlrun api we are in a separated thread, thus creating
            # a new event loop. in case of sdk, we are most likely in main thread, thus using the main event loop.
            try:
                event_loop = asyncio.get_event_loop()
            except RuntimeError:
                event_loop = asyncio.new_event_loop()
                asyncio.set_event_loop(event_loop)

            if not event_loop.is_running():
                event_loop.run_until_complete(async_push_callback())
            else:
                asyncio.run_coroutine_threadsafe(async_push_callback(), event_loop)

        # then push sync notifications
        if not mlrun.config.is_running_as_api():
            sync_push_callback()

    def _run_coroutine_in_jupyter_notebook(self, coroutine_method):
        """
        Execute a coroutine in a Jupyter Notebook environment.

        This function creates a new thread pool executor with a single thread and a new event loop.
        It sets the created event loop as the current event loop.
        Then, it submits the coroutine to the event loop and waits for its completion.

        This approach is used in Jupyter Notebook to ensure the proper execution of the event loop in a separate thread,
        allowing for the asynchronous push operation to be executed while the notebook is running.

        :param coroutine_method: The coroutine method to be executed.
        :return: The result of the executed coroutine.
        """
        thread_pool_executor = ThreadPoolExecutor(1)
        async_event_loop = asyncio.new_event_loop()
        thread_pool_executor.submit(asyncio.set_event_loop, async_event_loop).result()
        result = thread_pool_executor.submit(
            async_event_loop.run_until_complete, coroutine_method()
        ).result()
        return result


class NotificationPusher(_NotificationPusherBase):
    messages = {
        "completed": "{resource} completed",
        "error": "{resource} failed",
        "aborted": "{resource} aborted",
        "running": "{resource} started",
    }

    def __init__(
        self,
        runs: typing.Union[mlrun.lists.RunList, list],
        default_params: typing.Optional[dict] = None,
    ):
        self._runs = runs
        self._default_params = default_params or {}
        self._sync_notifications: list[
            tuple[
                base.NotificationBase, mlrun.model.RunObject, mlrun.model.Notification
            ]
        ] = []
        self._async_notifications: list[
            tuple[
                base.NotificationBase, mlrun.model.RunObject, mlrun.model.Notification
            ]
        ] = []

        for run in self._runs:
            try:
                self._process_run(run)
            except Exception as exc:
                logger.warning(
                    "Failed to process run",
                    run_uid=run.metadata.uid,
                    error=mlrun.errors.err_to_str(exc),
                )

    def _process_run(self, run):
        if isinstance(run, dict):
            run = mlrun.model.RunObject.from_dict(run)

        for notification in run.spec.notifications:
            try:
                self._process_notification(notification, run)
            except Exception as exc:
                logger.warning(
                    "Failed to process notification",
                    run_uid=run.metadata.uid,
                    notification=notification,
                    error=mlrun.errors.err_to_str(exc),
                )

    def _process_notification(self, notification_object, run):
        notification_object.status = run.status.notifications.get(
            notification_object.name, {}
        ).get(
            "status",
            mlrun.common.schemas.NotificationStatus.PENDING,
        )
        if self._should_notify(run, notification_object):
            notification = self._load_notification(notification_object)
            if notification.is_async:
                self._async_notifications.append(
                    (notification, run, notification_object)
                )
            else:
                self._sync_notifications.append(
                    (notification, run, notification_object)
                )

    def push(self, sync_push_callback=None, async_push_callback=None):
        """
        Asynchronously push notifications for all runs in the initialized runs list (if they should be pushed).
        When running from a sync environment, the notifications will be pushed asynchronously however the function will
        wait for all notifications to be pushed before returning.
        """

        if not len(self._sync_notifications) and not len(self._async_notifications):
            return

        def sync_push():
            for notification_data in self._sync_notifications:
                try:
                    self._push_notification_sync(
                        notification_data[0],
                        notification_data[1],
                        notification_data[2],
                    )
                except Exception as exc:
                    logger.warning(
                        "Failed to push notification sync",
                        error=mlrun.errors.err_to_str(exc),
                    )

        async def async_push():
            tasks = []
            for notification_data in self._async_notifications:
                tasks.append(
                    self._push_notification_async(
                        notification_data[0],
                        notification_data[1],
                        notification_data[2],
                    )
                )

            # return exceptions to "best-effort" fire all notifications
            results = await asyncio.gather(*tasks, return_exceptions=True)
            for result in results:
                if isinstance(result, Exception):
                    logger.warning(
                        "Failed to push notification async",
                        error=mlrun.errors.err_to_str(result),
                        traceback=traceback.format_exception(
                            result,
                            value=result,
                            tb=result.__traceback__,
                        ),
                    )

        logger.debug(
            "Pushing notifications",
            notifications_amount=len(self._sync_notifications)
            + len(self._async_notifications),
        )
        sync_push_callback = sync_push_callback or sync_push
        async_push_callback = async_push_callback or async_push
        self._push(sync_push_callback, async_push_callback)

    @staticmethod
    def _should_notify(
        run: mlrun.model.RunObject,
        notification: mlrun.model.Notification,
    ) -> bool:
        when_states = notification.when
        run_state = run.state()

        # if the notification isn't pending, don't push it
        if (
            notification.status
            and notification.status != mlrun.common.schemas.NotificationStatus.PENDING
        ):
            return False

        # if at least one condition is met, notify
        for when_state in when_states:
            if when_state == run_state:
                if (
                    run_state == runtimes_constants.RunStates.completed
                    and evaluate_condition_in_separate_process(
                        notification.condition,
                        context={
                            "run": run.to_dict(),
                            "notification": notification.to_dict(),
                        },
                    )
                ) or run_state in [
                    runtimes_constants.RunStates.error,
                    runtimes_constants.RunStates.aborted,
                    runtimes_constants.RunStates.running,
                ]:
                    return True

        return False

    def _load_notification(
        self, notification_object: mlrun.model.Notification
    ) -> base.NotificationBase:
        name = notification_object.name
        notification_type = notification_module.NotificationTypes(
            notification_object.kind or notification_module.NotificationTypes.console
        )
        params = {}
        params.update(notification_object.secret_params or {})
        params.update(notification_object.params or {})
        default_params = self._default_params.get(notification_type.value, {})
        notification = notification_type.get_notification()(
            name, params, default_params
        )
        logger.debug(
            "Loaded notification", notification=name, type=notification_type.value
        )
        return notification

    def _prepare_notification_args(
        self, run: mlrun.model.RunObject, notification_object: mlrun.model.Notification
    ):
        custom_message = (
            f": {notification_object.message}" if notification_object.message else ""
        )
        resource = "Run"
        runs = [run.to_dict()]

        if mlrun_constants.MLRunInternalLabels.workflow in run.metadata.labels:
            resource = mlrun_constants.MLRunInternalLabels.workflow
            custom_message = (
                f" (workflow: {run.metadata.labels['workflow']}){custom_message}"
            )
            project = run.metadata.project
            workflow_id = run.status.results.get("workflow_id", None)
            db = mlrun.get_run_db()
            runs.extend(Workflow.get_workflow_steps(db, workflow_id, project))

        message = (
            self.messages.get(run.state(), "").format(resource=resource)
            + f" in project {run.metadata.project}"
            + custom_message
        )

        retry_count = run.status.retry_count or 0
        max_retries = (run.spec.retry.count or 0) if run.spec.retry else 0

        # If any retries were attempted, include retry info in the final notification message.
        # This is only shown when the final notification is sent (after success or final failure)
        if retry_count > 0:
            message += f"\nRetries attempted: {retry_count}"
            if (
                run.state() == runtimes_constants.RunStates.error
                and retry_count >= max_retries
            ):
                message += (
                    "\nRetry limit reached - run has failed after all retry attempts."
                )

        severity = (
            notification_object.severity
            or mlrun.common.schemas.NotificationSeverity.INFO
        )
        return message, severity, runs

    def _push_notification_sync(
        self,
        notification: base.NotificationBase,
        run: mlrun.model.RunObject,
        notification_object: mlrun.model.Notification,
    ):
        message, severity, runs = self._prepare_notification_args(
            run, notification_object
        )

        logger.debug(
            "Pushing sync notification",
            notification=sanitize_notification(notification_object.to_dict()),
            run_uid=run.metadata.uid,
        )
        update_notification_status_kwargs = {
            "run_uid": run.metadata.uid,
            "project": run.metadata.project,
            "notification": notification_object,
            "status": mlrun.common.schemas.NotificationStatus.SENT,
            "run_state": run.state(),
        }
        try:
            notification.push(message, severity, runs)
            logger.debug(
                "Notification sent successfully",
                notification=sanitize_notification(notification_object.to_dict()),
                run_uid=run.metadata.uid,
            )
            update_notification_status_kwargs["sent_time"] = datetime.datetime.now(
                tz=datetime.timezone.utc
            )
        except Exception as exc:
            logger.warning(
                "Failed to send or update notification",
                notification=sanitize_notification(notification_object.to_dict()),
                run_uid=run.metadata.uid,
                exc=mlrun.errors.err_to_str(exc),
                traceback=traceback.format_exc(),
            )
            update_notification_status_kwargs["reason"] = f"Exception error: {str(exc)}"
            update_notification_status_kwargs["status"] = (
                mlrun.common.schemas.NotificationStatus.ERROR
            )
            raise exc
        finally:
            self._update_notification_status(
                **update_notification_status_kwargs,
            )

    async def _push_notification_async(
        self,
        notification: base.NotificationBase,
        run: mlrun.model.RunObject,
        notification_object: mlrun.model.Notification,
    ):
        message, severity, runs = self._prepare_notification_args(
            run, notification_object
        )

        logger.debug(
            "Pushing async notification",
            notification=sanitize_notification(notification_object.to_dict()),
            run_uid=run.metadata.uid,
        )
        update_notification_status_kwargs = {
            "run_uid": run.metadata.uid,
            "project": run.metadata.project,
            "notification": notification_object,
            "run_state": run.state(),
            "status": mlrun.common.schemas.NotificationStatus.SENT,
        }
        try:
            await notification.push(message, severity, runs)
            logger.debug(
                "Notification sent successfully",
                notification=sanitize_notification(notification_object.to_dict()),
                run_uid=run.metadata.uid,
            )
            update_notification_status_kwargs["sent_time"] = datetime.datetime.now(
                tz=datetime.timezone.utc
            )

        except Exception as exc:
            logger.warning(
                "Failed to send or update async notification",
                notification=sanitize_notification(notification_object.to_dict()),
                run_uid=run.metadata.uid,
                exc=mlrun.errors.err_to_str(exc),
                traceback=traceback.format_exc(),
            )
            update_notification_status_kwargs["reason"] = f"Exception error: {str(exc)}"
            update_notification_status_kwargs["status"] = (
                mlrun.common.schemas.NotificationStatus.ERROR
            )
            raise exc
        finally:
            await mlrun.utils.helpers.run_in_threadpool(
                self._update_notification_status,
                **update_notification_status_kwargs,
            )

    @staticmethod
    def _update_notification_status(
        run_uid: str,
        project: str,
        notification: mlrun.model.Notification,
        run_state: runtimes_constants.RunStates,
        status: typing.Optional[str] = None,
        sent_time: typing.Optional[datetime.datetime] = None,
        reason: typing.Optional[str] = None,
    ):
        # Skip update the notification state if the following conditions are met:
        # 1. the run is not in a terminal state
        # 2. the when contains only one state (which is the current state)
        # Skip updating because currently each notification has only one row in the db, even if it has multiple when.
        # This means that if the notification is updated to sent for running state for example, it will not send for
        # The terminal state
        # TODO: Change this behavior after implementing ML-8723
        if (
            run_state not in runtimes_constants.RunStates.terminal_states()
            and len(notification.when) > 1
        ):
            logger.debug(
                "Skip updating notification status - run not in terminal state",
                run_uid=run_uid,
                state=run_state,
            )
            return

        db = mlrun.get_run_db()
        notification.status = status or notification.status
        notification.sent_time = sent_time or notification.sent_time

        # fill reason only if failed
        if notification.status == mlrun.common.schemas.NotificationStatus.ERROR:
            notification.reason = reason or notification.reason

            # limit reason to a max of 255 characters (for db reasons)
            # but also for human readability reasons.
            notification.reason = notification.reason[:255]
        else:
            # empty out the reason if the notification is in a non-error state
            # in case a retry would kick in (when such mechanism would be implemented)
            notification.reason = None

        # There is no need to mask the secret_params as the secrets are already loaded
        db.store_run_notifications(
            [notification],
            run_uid,
            project,
            mask_params=False,
        )


class CustomNotificationPusher(_NotificationPusherBase):
    def __init__(self, notification_types: typing.Optional[list[str]] = None):
        notifications = {
            notification_type: notification_module.NotificationTypes(
                notification_type
            ).get_notification()()
            for notification_type in notification_types
        }
        self._sync_notifications = {
            notification_type: notification
            for notification_type, notification in notifications.items()
            if not notification.is_async
        }
        self._async_notifications = {
            notification_type: notification
            for notification_type, notification in notifications.items()
            if notification.is_async
        }
        self._server_notifications = []

    @property
    def notifications(self):
        notifications = self._sync_notifications.copy()
        notifications.update(self._async_notifications)
        return notifications

    @property
    def server_notifications(self):
        return self._server_notifications

    def push(
        self,
        message: str,
        severity: typing.Union[
            mlrun.common.schemas.NotificationSeverity, str
        ] = mlrun.common.schemas.NotificationSeverity.INFO,
        runs: typing.Union[mlrun.lists.RunList, list] = None,
        custom_html: typing.Optional[str] = None,
    ):
        def sync_push():
            for notification_type, notification in self._sync_notifications.items():
                if self.should_push_notification(notification_type):
                    notification.push(message, severity, runs, custom_html)

        async def async_push():
            tasks = []
            for notification_type, notification in self._async_notifications.items():
                if self.should_push_notification(notification_type):
                    tasks.append(
                        notification.push(message, severity, runs, custom_html)
                    )
            # return exceptions to "best-effort" fire all notifications
            await asyncio.gather(*tasks, return_exceptions=True)

        self._push(sync_push, async_push)

    def add_notification(
        self,
        notification_type: str,
        params: typing.Optional[dict[str, str]] = None,
        name: typing.Optional[str] = None,
        message: typing.Optional[str] = None,
        severity: mlrun.common.schemas.notification.NotificationSeverity = (
            mlrun.common.schemas.notification.NotificationSeverity.INFO
        ),
        when: typing.Optional[list[str]] = None,
        condition: typing.Optional[str] = None,
        secret_params: typing.Optional[dict[str, str]] = None,
    ):
        if notification_type not in [
            notification_module.NotificationTypes.console,
            notification_module.NotificationTypes.ipython,
        ]:
            # We want that only the console and ipython notifications will be notified by the client.
            # The rest of the notifications will be notified by the BE.
            self._server_notifications.append(
                mlrun.model.Notification(
                    kind=notification_type,
                    name=name,
                    message=message,
                    severity=severity,
                    when=when or runtimes_constants.RunStates.notification_states(),
                    params=params,
                    secret_params=secret_params,
                )
            )
            return

        if notification_type in self._async_notifications:
            self._async_notifications[notification_type].load_notification(params)
        elif notification_type in self._sync_notifications:
            self._sync_notifications[notification_type].load_notification(params)
        else:
            notification = notification_module.NotificationTypes(
                notification_type
            ).get_notification()(
                params=params,
            )
            if notification.is_async:
                self._async_notifications[notification_type] = notification
            else:
                self._sync_notifications[notification_type] = notification

    def remove_notification(self, notification_type: str):
        if notification_type in self._async_notifications:
            del self._async_notifications[notification_type]

        elif notification_type in self._sync_notifications:
            del self._sync_notifications[notification_type]

        else:
            logger.warning(f"No notification of type {notification_type} in project")

    def edit_notification(
        self, notification_type: str, params: typing.Optional[dict[str, str]] = None
    ):
        self.remove_notification(notification_type)
        self.add_notification(notification_type, params)

    def should_push_notification(self, notification_type):
        notifications = {}
        notifications.update(self._sync_notifications)
        notifications.update(self._async_notifications)
        notification = notifications.get(notification_type)
        if not notification or not notification.active:
            return False

        # get notification's inverse dependencies, and only push the notification if
        # none of its inverse dependencies are being sent
        inverse_dependencies = notification_module.NotificationTypes(
            notification_type
        ).inverse_dependencies()
        for inverse_dependency in inverse_dependencies:
            inverse_dependency_notification = notifications.get(inverse_dependency)
            if (
                inverse_dependency_notification
                and inverse_dependency_notification.active
            ):
                return False

        return True

    def push_pipeline_start_message(
        self,
        project: str,
        pipeline_id: typing.Optional[str] = None,
    ):
        db = mlrun.get_run_db()
        db.push_run_notifications(pipeline_id, project)

    def push_pipeline_start_message_from_client(
        self,
        project: str,
        commit_id: typing.Optional[str] = None,
        pipeline_id: typing.Optional[str] = None,
        has_workflow_url: bool = False,
    ):
        html, message = self.generate_start_message(
            commit_id, has_workflow_url, pipeline_id, project
        )
        self.push(message, "info", custom_html=html)

    def push_pipeline_run_results(
        self,
        runs: typing.Union[mlrun.lists.RunList, list],
        push_all: bool = False,
        state: typing.Optional[str] = None,
    ):
        """
        push a structured table with run results to notification targets

        :param runs:  list if run objects (RunObject)
        :param push_all: push all notifications (including already notified runs)
        :param state: final run state
        """
        had_errors = 0
        runs_list = []
        for run in runs:
            notified = getattr(run, "_notified", False)
            if not notified or push_all:
                if run.status.state == "error":
                    had_errors += 1
                runs_list.append(run.to_dict())
                run._notified = True

        text = "Pipeline run finished"
        if had_errors:
            text += f" with {had_errors} errors"
        if state:
            text += f", state={state}"
        self.push(text, "info", runs=runs_list)

    def generate_start_message(
        self, commit_id=None, has_workflow_url=None, pipeline_id=None, project=None
    ):
        message = f"Workflow started in project {project}"
        if pipeline_id:
            message += f" id={pipeline_id}"
        commit_id = (
            commit_id or os.environ.get("GITHUB_SHA") or os.environ.get("CI_COMMIT_SHA")
        )
        if commit_id:
            message += f", commit={commit_id}"
        if has_workflow_url:
            url = mlrun.utils.helpers.get_workflow_url(project, pipeline_id)
        else:
            url = mlrun.utils.helpers.get_runs_url(project)
        html = ""
        if url:
            html = (
                message
                + f'<div><a href="{url}" target="_blank">click here to view progress</a></div>'
            )
            message = message + f", check progress in {url}"
        return html, message


def sanitize_notification(notification_dict: dict):
    notification_dict.pop("secret_params", None)
    notification_dict.pop("message", None)
    notification_dict.pop("params", None)
    return notification_dict



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import typing

import mlrun.common.schemas
import mlrun.common.types

from .base import ObjectFormat


class ProjectFormat(ObjectFormat, mlrun.common.types.StrEnum):
    full = "full"
    name_only = "name_only"
    # minimal format removes large fields from the response (e.g. functions, workflows, artifacts)
    # and is used for faster response times (in the UI)
    minimal = "minimal"
    # internal - allowed only in follower mode, only for the leader for upgrade purposes
    leader = "leader"

    name_and_creation_time = "name_and_creation_time"

    @staticmethod
    def format_method(_format: str) -> typing.Optional[typing.Callable]:
        def _name_only(project: mlrun.common.schemas.Project) -> str:
            return project.metadata.name

        def _name_and_creation_time(
            project: mlrun.common.schemas.Project,
        ) -> tuple[str, datetime.datetime]:
            return project.metadata.name, project.metadata.created

        def _minimal(
            project: mlrun.common.schemas.Project,
        ) -> mlrun.common.schemas.Project:
            project.spec.functions = None
            project.spec.workflows = None
            project.spec.artifacts = None
            return project

        return {
            ProjectFormat.full: None,
            ProjectFormat.name_only: _name_only,
            ProjectFormat.minimal: _minimal,
            ProjectFormat.leader: None,
            ProjectFormat.name_and_creation_time: _name_and_creation_time,
        }[_format]
